import re
import warnings
from dataclasses import dataclass, field
from enum import Enum
from types import UnionType
from typing import Any, Literal, Type, Union

import pydantic as pyd
import pytest
from test_worker.main import EmptyStruct, IntStruct, MyGeneric

from tierkreis.builder import Const, Output, UnionConst, graph
from tierkreis.client.runtime_client import RuntimeClient
from tierkreis.core.opaque_model import OpaqueModel
from tierkreis.core.types import IncompatibleUnionType, TierkreisType, UnionTag
from tierkreis.core.values import (
    FloatValue,
    IntValue,
    StringValue,
    StructValue,
    TierkreisValue,
    ToPythonFailure,
    VariantValue,
    VecValue,
    option_none,
)


@pytest.mark.asyncio
async def test_union_types(bi, client: RuntimeClient) -> None:
    # test structs containing unions can be converted to variant values, sent
    # through worker functions that process them as unions, and converted back
    # to unions.
    iv = TierkreisValue.from_python(1)
    assert iv.to_python_union(float | int) == 1
    fv = TierkreisValue.from_python(2.3)
    assert fv.to_python_union(int | float) == 2.3

    @dataclass
    class StructWithUnion:
        # redefine struct with Union[] syntax here to test both
        # (alongside A | B syntax)
        x: Union[IntStruct, float]

    pn = bi["python_nodes"]
    a_py = StructWithUnion(1.0)
    b_py = StructWithUnion(IntStruct(2))

    @graph()
    def g() -> Output:
        return Output(
            a=pn.id_union_struct(Const(a_py)), b=pn.id_union_struct(Const(b_py))
        )

    out = await client.run_graph(g)
    a, b = out["a"], out["b"]

    assert isinstance(a, StructValue)
    assert isinstance(b, StructValue)
    a_var = a.values["x"]
    assert isinstance(a_var, VariantValue)
    assert a_var.tag == UnionTag.prefix + "float"
    b_var = b.values["x"]
    assert isinstance(b_var, VariantValue)
    assert b_var.tag == UnionTag.prefix + "int_struct"

    assert a.to_python(StructWithUnion) == a_py
    assert b.to_python(StructWithUnion) == b_py

    for bad_tag in ("invalid_tag", UnionTag.type_tag(IntStruct)):
        with pytest.raises(ToPythonFailure):
            v: VariantValue = VariantValue(bad_tag, TierkreisValue.from_python(1.0))
            struct: StructValue = StructValue({"x": v})
            _ = struct.to_python(StructWithUnion)

    @dataclass
    class AmbiguousUnion:
        x: Union[list[int], list[bool]]

    with pytest.raises(IncompatibleUnionType):
        TierkreisType.from_python(AmbiguousUnion)

    sig = await client.get_signature()

    @graph(type_check_sig=sig)
    def g2() -> Output:
        return Output(a=pn.id_union(UnionConst(1)), b=pn.id_union(UnionConst(1.0)))

    out = await client.run_graph(g2)
    a, b = out["a"], out["b"]
    assert a.to_python_union(int | float) == 1
    assert b.to_python_union(int | float) == 1.0


@pyd.dataclasses.dataclass(frozen=True, kw_only=True)
class ConstrainedField:
    foo: float
    bar: pyd.PositiveFloat | None = None
    points: tuple[float, ...]


@pyd.dataclasses.dataclass(frozen=True, kw_only=True)
class SimpleVariant(ConstrainedField):
    mean: float
    disc_name: Literal["Simple", "simple"] = field(default="Simple")


@pyd.dataclasses.dataclass(frozen=True, kw_only=True)
class ComplexVariant(ConstrainedField):
    mean: float
    variance: pyd.PositiveFloat
    foo: pyd.NonNegativeFloat
    disc_name: Literal["Complex"] = field(default="Complex", init=True)


class EnumExample(Enum):
    First = "First"
    Second = "Second"


class ContainsVariant(pyd.BaseModel):
    n: int
    enum: EnumExample
    variant: ComplexVariant | SimpleVariant = pyd.Field(discriminator="disc_name")


@pyd.dataclasses.dataclass(frozen=True, kw_only=True)
class NonInitPydField:
    xl: int = pyd.Field(default=0)
    non_init: int = field(default=0, init=False)


class Model(pyd.BaseModel):
    x: int
    y: NonInitPydField


class AnnotatedWithUnion(pyd.BaseModel):
    un: pyd.PositiveFloat | int


class AnnotatedInList(pyd.BaseModel):
    ul: list[pyd.PositiveFloat]


class OpaqueContainsVariant(OpaqueModel):
    n: int
    enum: EnumExample
    variant: ComplexVariant | SimpleVariant = pyd.Field(discriminator="disc_name")


def _tktype_from_py(py_type: UnionType | Type) -> TierkreisType:
    if isinstance(py_type, UnionType):
        return TierkreisType.from_python_union(py_type)
    return TierkreisType.from_python(py_type)


def _tkval_from_py(py_val: Any, py_type: UnionType | Type | None) -> TierkreisValue:
    if isinstance(py_type, UnionType):
        return TierkreisValue.from_python_union(py_val, py_type)
    return TierkreisValue.from_python(py_val, py_type)


def _tkval_to_py(tkv: TierkreisValue, py_type: UnionType | Type) -> Any:
    if isinstance(py_type, UnionType):
        return tkv.to_python_union(py_type)
    return tkv.to_python(py_type)


@pytest.mark.asyncio
async def test_pydantic_types(bi, client: RuntimeClient) -> None:
    constrained = ConstrainedField(foo=1, points=(1.2,))
    tk_constrained = StructValue(
        {
            "foo": FloatValue(1.0),
            "points": VecValue([FloatValue(1.2)]),
            "bar": option_none,
        }
    )

    cmplex = ComplexVariant(mean=-3.4, variance=4.5, **constrained.__dict__)
    tk_cmplex = StructValue(
        {
            "mean": FloatValue(-3.4),
            "variance": FloatValue(4.5),
            "disc_name": StringValue("Complex"),
            **tk_constrained.values,
        }
    )

    simple = SimpleVariant(mean=1.3, **constrained.__dict__)
    tk_simple = StructValue(
        {
            **tk_constrained.values,
            "mean": FloatValue(1.3),
            "disc_name": StringValue("Simple"),
        }
    )

    contains = ContainsVariant(variant=cmplex, n=4, enum=EnumExample.First)
    opaque_contains = OpaqueContainsVariant(variant=cmplex, n=4, enum=EnumExample.First)

    tk_first = VariantValue("First", StructValue({}))
    tk_contains = StructValue(
        {
            "variant": VariantValue("Complex", tk_cmplex),
            "n": IntValue(4),
            "enum": tk_first,
        }
    )

    contains_alt = ContainsVariant(variant=simple, n=4, enum=EnumExample.Second)
    tk_second = VariantValue("Second", StructValue({}))
    tk_contains_alt = StructValue(
        {
            "variant": VariantValue("Simple", tk_simple),
            "n": IntValue(4),
            "enum": tk_second,
        }
    )

    da = NonInitPydField()
    tk_da = StructValue(
        {
            "xl": IntValue(0),
            "non_init": IntValue(0),
        }
    )

    tk_model = StructValue(
        {
            "x": IntValue(1),
            "y": tk_da,
        }
    )

    tk_list_model = StructValue(
        {
            "ul": VecValue([FloatValue(1.2)]),
        }
    )

    tk_union_model = StructValue(
        {
            "un": VariantValue(UnionTag.type_tag(int), IntValue(1)),
        }
    )
    tk_opaque_contains = StructValue(
        {
            "__tk_opaque_opaque_contains_variant": StringValue(
                opaque_contains.model_dump_json()
            ),
        }
    )

    def mk_mygeneric_variant(typ: Type, val: TierkreisValue) -> StructValue:
        return StructValue({"x": VariantValue(UnionTag.type_tag(typ), val)})

    samples: list[tuple[Type | UnionType, Any, TierkreisValue]] = [
        (ConstrainedField, constrained, tk_constrained),
        (SimpleVariant, simple, tk_simple),
        (EnumExample, EnumExample.First, tk_first),
        (EnumExample, EnumExample.Second, tk_second),
        (ComplexVariant, cmplex, tk_cmplex),
        (ContainsVariant, contains, tk_contains),
        (ContainsVariant, contains_alt, tk_contains_alt),
        (NonInitPydField, da, tk_da),
        (Model, Model(x=1, y=da), tk_model),
        (AnnotatedInList, AnnotatedInList(ul=[1.2]), tk_list_model),
        (AnnotatedWithUnion, AnnotatedWithUnion(un=1), tk_union_model),
        (
            OpaqueContainsVariant,
            opaque_contains,
            tk_opaque_contains,
        ),
        (EnumExample | None, EnumExample.First, tk_first),
        # (None | EnumExample, EnumExample.First, tk_first), # fails, see test_reversed_union
        (EnumExample | None, None, option_none),
        (None | EnumExample, None, option_none),
    ]

    for py_type, py_val, expected_tk_val in samples:
        _ = _tktype_from_py(py_type)
        converted_tk_val = TierkreisValue.from_python(py_val)
        assert converted_tk_val == expected_tk_val
        assert _tkval_from_py(py_val, py_type) == expected_tk_val
        assert _tkval_to_py(converted_tk_val, py_type) == py_val

        pn = bi["python_nodes"]

        @graph()
        def g() -> Output:
            return Output(pn.id_py(Const(py_val)))

        out = (await client.run_graph(g))["value"]

        assert _tkval_to_py(out, py_type) == py_val


def test_union_over_arguments() -> None:
    values = [MyGeneric[int](x=3), MyGeneric[str](x="foo")]
    types = [
        t
        for base in [MyGeneric[int] | MyGeneric[str], MyGeneric[str] | MyGeneric[int]]
        for t in [base, base | None, None | base]
    ]

    for val in values:
        for typ in types:
            tk_val = TierkreisValue.from_python_union(val, typ)
            assert tk_val.to_python_union(typ) == val

            # Separately test value+typ embedded together in an outer MyGeneric
            typ2 = MyGeneric[typ]  # type: ignore
            val2 = typ2(x=val)
            tk_val2 = TierkreisValue.from_python(val2, typ2)
            assert tk_val2.to_python(typ2) == val2


@pytest.mark.xfail(reason="See issue #454")
def test_reversed_union_enum() -> None:
    # This should be an entry in test:pydantic_types, if it worked:
    # (None | EnumExample, EnumExample.First, tk_first)
    tk_first = VariantValue("First", StructValue({}))
    py_type, py_val, expected_tk_val = (None | EnumExample, EnumExample.First, tk_first)
    assert TierkreisValue.from_python(py_val) == expected_tk_val
    # Works the right way around:
    assert expected_tk_val.to_python_union(EnumExample | None) == py_val
    # But reverse the union and it does not (result is "None")
    assert expected_tk_val.to_python_union(py_type) == py_val


def test_pydantic_generic_union() -> None:
    # Similar to above but from_python with explicit 'typ'
    # avoids the need for pydantic validation
    val = MyGeneric(x=MyGeneric[float](x=3.2))
    typ = MyGeneric[MyGeneric[int] | MyGeneric[float]]
    tk_val = TierkreisValue.from_python(val, typ)
    assert tk_val.to_python(typ) == val


@pytest.mark.asyncio
async def test_return_union(bi, client: RuntimeClient) -> None:
    # test a worker function declared as returning a union type,
    # correctly converts the return value

    sig = await client.get_signature()
    pn = bi["python_nodes"]

    @graph(type_check_sig=sig)
    def g() -> Output:
        return Output(b=pn.zero_to_empty(x=Const(0)), a=pn.zero_to_empty(x=Const(1)))

    out = await client.run_graph(g)

    a, b = out["a"], out["b"]

    assert isinstance(a, VariantValue)
    assert a.tag == UnionTag.type_tag(IntStruct)
    assert isinstance(b, VariantValue)
    assert b.tag == UnionTag.type_tag(EmptyStruct)

    # to_python is tag-driven so ordering of union elements makes no difference
    assert a.to_python_union(IntStruct | EmptyStruct) == IntStruct(1)
    assert a.to_python_union(EmptyStruct | IntStruct) == IntStruct(1)
    assert b.to_python_union(IntStruct | EmptyStruct) == EmptyStruct()


# Make sure that the ordering of Union elements doesn't affect anything here
@pytest.mark.parametrize(
    "union_type", [IntStruct | EmptyStruct, EmptyStruct | IntStruct]
)
def test_warnings(union_type: Type) -> None:
    with warnings.catch_warnings(record=True) as ws:
        val = TierkreisValue.from_python(IntStruct(3), union_type)
    assert len(ws) == 0
    assert val == VariantValue(
        UnionTag.type_tag(IntStruct), StructValue({"y": IntValue(3)})
    )

    @dataclass
    class IntSubStruct(IntStruct):
        new_field: float

    with warnings.catch_warnings(record=True) as ws:
        val = TierkreisValue.from_python(IntSubStruct(13, 4.21), union_type)
    assert len(ws) == 1
    assert re.search("IntSubStruct.*beyond those in.*IntStruct", str(ws[0]))
    assert val == VariantValue(
        UnionTag.type_tag(IntStruct), StructValue({"y": IntValue(13)})
    )


@pytest.mark.asyncio
async def test_union_generic_basemodel(sig, bi, client: RuntimeClient) -> None:
    msg = "Hello World!"

    @graph(type_check_sig=sig)
    def main() -> Output:
        res = bi["python_nodes"].echo_union(Const(msg))
        return Output(res)

    res = await client.run_graph(main)
    assert res.pop("value").to_python_union(MyGeneric[str] | int) == MyGeneric[str](
        x=msg
    )
    assert res == {}  # No other outputs besides x



"""Tierkreis python tools"""

# remove with betterproto update
from betterproto import (
    PACKED_TYPES,
    TYPE_BYTES,
    Message,
    _preprocess_single,
    _serialize_single,
)

# only import elements from core package
from tierkreis.core.graphviz import render_graph, tierkreis_to_graphviz
from tierkreis.core.tierkreis_graph import TierkreisGraph, TierkreisType, TierkreisValue

from ._version import __version__



"""Errors generated by type inference system."""

from dataclasses import dataclass, field, replace
from typing import TYPE_CHECKING, Any, Optional, cast

import betterproto

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
import tierkreis.core.protos.tierkreis.v1alpha1.signature as ps
from tierkreis.core.function import FunctionName
from tierkreis.core.tierkreis_graph import (
    BoxNode,
    ConstNode,
    FunctionNode,
    GraphValue,
    MatchNode,
    NodeRef,
    TagNode,
)
from tierkreis.core.types import GraphType, Row, TierkreisType

if TYPE_CHECKING:
    from tierkreis.core.tierkreis_graph import TierkreisGraph, TierkreisNode


def _node_port_str(idx: int, port: str, g: "TierkreisGraph") -> str:
    return f'{_node_str(idx, g[idx])[0]}["{port}"]'


def _node_str(idx: int, tkn: "TierkreisNode") -> tuple[str, Optional["TierkreisGraph"]]:
    # if node contains another graph return it as second value
    s = str(idx)
    g = None
    if isinstance(tkn, FunctionNode):
        s += f", Function({str(tkn.function_name)})"
    elif isinstance(tkn, BoxNode):
        g = tkn.graph
        s += ", Box"
        if tkn.location:
            s += "@"
            s += "/".join(tkn.location.location)
        if tkn.graph.name:
            s += f"({tkn.graph.name})"
    elif isinstance(tkn, TagNode):
        s += f", Tag({tkn.tag_name})"

    elif isinstance(tkn, MatchNode):
        s += ", Match"
    elif isinstance(tkn, ConstNode):
        val = tkn.value
        if isinstance(val, GraphValue) and val.value.name:
            val_str = f"Graph({val.value.name})"
            g = val.value
        else:
            val_str = val.viz_str()
        s += f", Const({val_str})"
    return f"NodeRef({s})", g


_NODE_IDX_LOCS = (
    "input",
    "output",
    "pair_first",
    "pair_second",
    "map_key",
    "map_value",
)


@dataclass(frozen=True)
class TierkreisTypeError(Exception):
    """Base class for all type errors."""

    proto_err: ps.TierkreisTypeError
    graph: "TierkreisGraph"
    dbg: dict[NodeRef, str] = field(default_factory=dict)

    def __str__(self) -> str:
        # default type error string
        return f"{self.proto_err.variant}\n\nIn:{self.location_str()}"

    @classmethod
    def from_proto(
        cls,
        proto: ps.TierkreisTypeError,
        g: "TierkreisGraph",
        debug: dict[NodeRef, str],
    ) -> "TierkreisTypeError":
        name, _ = betterproto.which_one_of(proto.variant, "error")
        cls_map = {
            "unify": UnifyError,
            "unknown_function": UnknownFunction,
        }

        # TODO more error variant special casing
        return cls_map.get(name, cls)(proto, g, debug)

    def _get_dbg_str(
        self, name: str, out_type: Any, g: "TierkreisGraph"
    ) -> Optional[str]:
        if name not in ("node_idx", "input", "output"):
            return ""
        idx = 0 if name == "input" else 1 if name == "output" else cast(int, out_type)
        return self.dbg.get(NodeRef(idx, g))

    def location_str(self) -> str:
        locs = []
        dbgs = []
        current_g = self.graph
        for location in self.proto_err.location:
            name, out_type = betterproto.which_one_of(location, "location")
            if dbg_str := self._get_dbg_str(name, out_type, current_g):
                # s += f"\n{dbg_str}"
                dbgs.append(dbg_str)
            if name == "root":
                continue
            elif name in _NODE_IDX_LOCS:
                locs.append(name)
            elif name == "vec_index":
                locs.append(f"Vec index[{out_type}]")
            elif name == "struct_field":
                locs.append(f"StructField({out_type})")
            elif name == "input_value":
                locs.append(f"InputValue({out_type})")
            elif name == "edge":
                edge = cast(pg.Edge, out_type)
                src = _node_port_str(edge.node_from, edge.port_from, current_g)
                tgt = _node_port_str(edge.node_to, edge.port_to, current_g)
                locs.append(f"Edge({src} -> {tgt})")
            elif name == "node_idx":
                idx = cast(int, out_type)
                s_n, next_g = _node_str(idx, current_g[idx])
                current_g = next_g or current_g
                locs.append(s_n)
            else:
                locs.append(str(out_type))

        s = "In: "
        s += "/".join(locs)
        if dbgs:
            s += "\nDebug Info:\n" + "\n".join(dbgs)

        return s

    def to_proto(self) -> ps.TierkreisTypeError:
        return self.proto_err


def _row_diff_hint(expected_r: Row, found_r: Row, name: str) -> str:
    hintstr = ""
    if expected_r.rest is None and (
        inter := set(found_r.content).difference(expected_r.content)
    ):
        hintstr += f"There are extra {name}: {inter}.\n"
    if found_r.rest is None and (
        inter := set(expected_r.content).difference(found_r.content)
    ):
        hintstr += f"There are missing {name}: {inter}.\n"

    return hintstr


class UnifyError(TierkreisTypeError):
    """Type inference system failed to unify two types. A particular type was
    expected based on other constraints, but a different type was found.
    """

    def __str__(self) -> str:
        expected = TierkreisType.from_proto(self.proto_err.variant.unify.expected)
        found = TierkreisType.from_proto(self.proto_err.variant.unify.found)
        s = f"Expected type {expected} but found {found}"

        s += "\n"
        s += self.location_str()

        if isinstance(expected, GraphType) and isinstance(found, GraphType):
            hintstr = _row_diff_hint(expected.inputs, found.inputs, "inputs")
            hintstr += _row_diff_hint(expected.outputs, found.outputs, "outputs")
            if hintstr:
                s += f"\nHint: {hintstr}"
        return s


class UnknownFunction(TierkreisTypeError):
    """An unknown function name was used - no connected worker has this function
    in the signature.
    """

    def __str__(self) -> str:
        fnam = FunctionName.from_proto(self.proto_err.variant.unknown_function)
        return f"Unknown function: {str(fnam)}\n{self.location_str()}"


@dataclass()
class TierkreisTypeErrors(Exception):
    """List of all type errors in a graph."""

    errors: list[TierkreisTypeError]

    @classmethod
    def from_proto(
        cls, proto: ps.TypeErrors, g: "TierkreisGraph"
    ) -> "TierkreisTypeErrors":
        return cls(
            errors=[
                TierkreisTypeError.from_proto(error, g, {}) for error in proto.errors
            ],
        )

    def to_proto(self) -> ps.TypeErrors:
        return ps.TypeErrors([e.to_proto() for e in self.errors])

    def __str__(self) -> str:
        separator = "\n\n" + ("─" * 80) + "\n\n"
        return separator.join(map(str, self.errors))

    def __getitem__(self, index: int) -> TierkreisTypeError:
        return self.errors[index]

    def __len__(self) -> int:
        return len(self.errors)

    def with_debug(self, dbg: dict[NodeRef, str]) -> "TierkreisTypeErrors":
        """Add debug information to all errors."""
        self.errors = [replace(e, dbg=dbg) for e in self.errors]
        return self



"""Implementation of builtin namespace in python."""

import asyncio
from copy import deepcopy
from dataclasses import dataclass
from typing import Generic, TypeVar, cast

from tierkreis.core.protos.tierkreis.v1alpha1.graph import (
    Constraint,
    Empty,
    GraphType,
    Kind,
    PartitionConstraint,
    RowType,
    StructType,
    Type,
    TypeScheme,
    TypeSchemeVar,
)
from tierkreis.core.protos.tierkreis.v1alpha1.signature import FunctionDeclaration
from tierkreis.core.python import RuntimeGraph
from tierkreis.core.tierkreis_graph import GraphValue, IncomingWireType, TierkreisGraph
from tierkreis.core.types import StarKind, TierkreisPair, UnpackRow
from tierkreis.core.utils import map_vals
from tierkreis.core.values import StructValue
from tierkreis.worker.namespace import Function, Namespace

namespace = Namespace()

a = TypeVar("a")
b = TypeVar("b")
val = TypeVar("val")
key = TypeVar("key")


@namespace.function(name="and")
async def _and(a: bool, b: bool) -> bool:
    """Check a and b are true; a && b"""
    return a and b


@dataclass
class CopyOut(Generic[a], UnpackRow):
    value_0: a
    value_1: a


@namespace.function(type_vars={"a": StarKind()})
async def copy(value: a) -> CopyOut[a]:
    "Copies its input value to each of its two outputs."
    return CopyOut(value, deepcopy(value))


@dataclass
class EmptyOut(Generic[a], UnpackRow):
    pass


@namespace.function(type_vars={"a": StarKind()})
async def discard(value: a) -> EmptyOut[a]:
    """Ignores its input value, has no outputs."""
    # things won't actually be deleted until python decides to
    return EmptyOut()


@dataclass
class EqOut(UnpackRow):
    result: bool


@namespace.function(type_vars={"val": StarKind()})
async def eq(value_0: val, value_1: val) -> EqOut:
    """Check two input values of the same type for equality, \
producing a boolean."""
    return EqOut(value_0 == value_1)


async def _eval(_client, _stack, _x: StructValue) -> StructValue:
    # implemented as part of runtime
    raise NotImplementedError


namespace.functions["eval"] = Function(
    run=_eval,
    declaration=FunctionDeclaration(
        type_scheme=TypeScheme(
            variables=[
                TypeSchemeVar(name="in", kind=Kind(row=Empty())),
                TypeSchemeVar(name="out", kind=Kind(row=Empty())),
            ],
            body=Type(
                graph=GraphType(
                    inputs=RowType(
                        content={
                            "thunk": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="in"),
                                    outputs=RowType(rest="out"),
                                )
                            )
                        },
                        rest="in",
                    ),
                    outputs=RowType(rest="out"),
                )
            ),
        ),
        description="Evaluates the graph on the `thunk` input with other inputs "
        "matching the graph inputs, producing outputs matching those of the graph.",
        input_order=["thunk"],
    ),
)


@namespace.function()
async def fadd(a: float, b: float) -> float:
    """Add floats a and b together; a + b"""
    return a + b


@namespace.function()
async def fdiv(a: float, b: float) -> float:
    """float division of a by b; a / b"""
    return a / b


@namespace.function()
async def fgeq(a: float, b: float) -> bool:
    """Check if a is greater than or equal to b; a >= b"""
    return a >= b


@namespace.function()
async def fgt(a: float, b: float) -> bool:
    """Check if a is greater than b; a > b"""
    return a > b


@namespace.function()
async def fleq(a: float, b: float) -> bool:
    """Check if a is less than or equal to b; a <= b"""
    return a <= b


@namespace.function()
async def float_to_int(float: float) -> int:
    """Convert a float to an integer"""
    return int(float)


@namespace.function()
async def flt(a: float, b: float) -> bool:
    """Check if a is less than b; a < b"""
    return a < b


@namespace.function()
async def fmod(a: float, b: float) -> float:
    """Modulo of a by b; a % b"""
    return a % b


@namespace.function()
async def fmul(a: float, b: float) -> float:
    """Multiply floats a and b together; a * b"""
    return a * b


@namespace.function()
async def fpow(a: float, b: float) -> float:
    """Exponentiate a by power b; a ^ b"""
    return a**b


@namespace.function()
async def fsub(a: float, b: float) -> float:
    """Subtract float b from float a; a - b"""
    return a - b


@namespace.function()
async def iadd(a: int, b: int) -> int:
    """Add integers a and b together; a + b"""
    return a + b


@namespace.function("id", type_vars={"a": StarKind()})
async def _id(value: a) -> a:
    """Passes a single, arbitrary, value from input to output."""
    return value


@namespace.function(type_vars={"a": StarKind()})
async def sleep(value: a, delay_secs: float) -> a:
    """Identity function with an asynchronous delay input in seconds."""
    await asyncio.sleep(delay_secs)
    return value


@namespace.function()
async def idiv(a: int, b: int) -> int:
    """Integer division of a by b; a / b"""
    return a // b


@namespace.function()
async def igeq(a: int, b: int) -> bool:
    """Check if a is greater than or equal to b; a >= b"""
    return a >= b


@namespace.function()
async def igt(a: int, b: int) -> bool:
    """Check if a is greater than b; a > b"""
    return a > b


@namespace.function()
async def ileq(a: int, b: int) -> bool:
    """Check if a is less than or equal to b; a <= b"""
    return a <= b


@namespace.function()
async def ilt(a: int, b: int) -> bool:
    """Check if a is less than b; a < b"""
    return a < b


@namespace.function()
async def imod(a: int, b: int) -> int:
    """Modulo of a by b; a % b"""
    return a % b


@namespace.function()
async def imul(a: int, b: int) -> int:
    """Multiply integers a and b together; a * b"""
    return a * b


@dataclass()
class _InsertOut(Generic[a, b], UnpackRow):
    map: dict[a, b]


@namespace.function(type_vars={"a": StarKind(), "b": StarKind()})
async def insert_key(map: dict[a, b], key: a, val: b) -> _InsertOut[a, b]:
    """Transforms an input `Map` into an output by adding an \
entry (or replacing an existing one) for an input key and value."""
    map[key] = val
    return _InsertOut(map)


@namespace.function()
async def int_to_float(int: int) -> float:
    """Convert an integer to a float"""
    return float(int)


@namespace.function()
async def ipow(a: int, b: int) -> int:
    """Exponentiate a by power b; a ^ b. b must be a positive integer."""
    return a**b


@namespace.function()
async def isub(a: int, b: int) -> int:
    """Subtract integer b from integer a; a - b"""
    return a - b


async def _loop(_client, _stack, _x: StructValue) -> StructValue:
    # implemented as part of runtime
    raise NotImplementedError


namespace.functions["loop"] = Function(
    run=_loop,
    declaration=FunctionDeclaration(
        type_scheme=TypeScheme(
            variables=[
                TypeSchemeVar(name="loop_var", kind=Kind(star=Empty())),
                TypeSchemeVar(name="result", kind=Kind(star=Empty())),
            ],
            body=Type(
                graph=GraphType(
                    inputs=RowType(
                        content={
                            "body": Type(
                                graph=GraphType(
                                    inputs=RowType(
                                        content={"value": Type(var="loop_var")}
                                    ),
                                    outputs=RowType(
                                        content={
                                            "value": Type(
                                                variant=RowType(
                                                    content={
                                                        "break": Type(var="result"),
                                                        "continue": Type(
                                                            var="loop_var"
                                                        ),
                                                    }
                                                )
                                            )
                                        }
                                    ),
                                )
                            ),
                            "value": Type(var="loop_var"),
                        }
                    ),
                    outputs=RowType(content={"value": Type(var="result")}),
                )
            ),
        ),
        description="Repeatedly applies a `Graph` input while it produces \
a `Variant` tagged `continue`, i.e. until it returns a value tagged `break`, \
and then returns that value.",
        input_order=["body", "value"],
        output_order=["value"],
    ),
)


@dataclass
class _MakePairOut(Generic[a, b], UnpackRow):
    pair: TierkreisPair[a, b]


@namespace.function(type_vars={"a": StarKind(), "b": StarKind()})
async def make_pair(first: a, second: b) -> _MakePairOut[a, b]:
    """Makes an output `Pair` from two separate inputs."""
    return _MakePairOut(TierkreisPair(first, second))


async def make_struct(_client, _stack, ins: StructValue) -> StructValue:
    """Construct a struct from incoming ports."""
    return StructValue({"struct": ins})


namespace.functions["make_struct"] = Function(
    run=make_struct,
    declaration=FunctionDeclaration(
        type_scheme=TypeScheme(
            variables=[TypeSchemeVar(name="fields", kind=Kind(row=Empty()))],
            body=Type(
                graph=GraphType(
                    inputs=RowType(rest="fields"),
                    outputs=RowType(
                        content={
                            "struct": Type(
                                struct=StructType(shape=RowType(rest="fields"))
                            )
                        }
                    ),
                )
            ),
        ),
        description="Takes any number of inputs and produces a single `Struct` "
        "output with a field for each input, field names matching input ports.",
        output_order=["struct"],
    ),
)


@namespace.function(type_vars={"val": StarKind()})
async def neq(value_0: val, value_1: val) -> EqOut:
    """Check two values are not equal."""
    return EqOut(value_0 != value_1)


@namespace.function("or")
async def _or(a: bool, b: bool) -> bool:
    """Check a or b are true; a || b"""
    return a or b


async def _partial(_client, _stack, inputs: StructValue) -> StructValue:
    invals = inputs.values
    thunk = deepcopy(cast(GraphValue, invals.pop("thunk")).value)
    newg = TierkreisGraph()
    rest_inputs = [port for port in thunk.inputs() if port not in invals]
    inports = map_vals(invals, lambda x: newg.add_const(x)["value"])
    inports.update({port: newg.input[port] for port in rest_inputs})
    outs = newg.insert_graph(thunk, **inports)
    newg.set_outputs(**outs)
    return StructValue({"value": GraphValue(newg)})


namespace.functions["partial"] = Function(
    run=_partial,
    declaration=FunctionDeclaration(
        type_scheme=TypeScheme(
            variables=[
                TypeSchemeVar(name="in_given", kind=Kind(row=Empty())),
                TypeSchemeVar(name="in_rest", kind=Kind(row=Empty())),
                TypeSchemeVar(name="in", kind=Kind(row=Empty())),
                TypeSchemeVar(name="outv", kind=Kind(row=Empty())),
            ],
            constraints=[
                Constraint(
                    partition=PartitionConstraint(
                        left=Type(var="in_given"),
                        right=Type(var="in_rest"),
                        union=Type(var="in"),
                    )
                )
            ],
            body=Type(
                graph=GraphType(
                    inputs=RowType(
                        content={
                            "thunk": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="in"),
                                    outputs=RowType(rest="outv"),
                                )
                            )
                        },
                        rest="in_given",
                    ),
                    outputs=RowType(
                        content={
                            "value": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="in_rest"),
                                    outputs=RowType(rest="outv"),
                                )
                            )
                        }
                    ),
                )
            ),
        ),
        description="Partial application;"
        " output a new graph with some input values injected as constants",
        input_order=["thunk"],
        output_order=["value"],
    ),
)


@dataclass
class _PopOut(Generic[a], UnpackRow):
    vec: list[a]
    item: a


@namespace.function(type_vars={"a": StarKind()})
async def pop(vec: list[a]) -> _PopOut[a]:
    """Pops the first element from an input `Vec`, returning said \
element separately from the remainder `Vec`. Fails at runtime if the input `Vec` \
is empty."""
    item = vec.pop()
    return _PopOut(vec, item)


@dataclass
class _PushOut(Generic[a], UnpackRow):
    vec: list[a]


@namespace.function(type_vars={"a": StarKind()})
async def push(vec: list[a], item: a) -> _PushOut[a]:
    """Adds an input element onto an input `Vec` to give an output `Vec`."""
    vec.append(item)
    return _PushOut(vec)


@dataclass
class _RemoveOut(Generic[a, b], UnpackRow):
    map: dict[a, b]
    val: b


@namespace.function(type_vars={"a": StarKind(), "b": StarKind()})
async def remove_key(map: dict[a, b], key: a) -> _RemoveOut[a, b]:
    """Remove a key (input) from a map (input), return the map and value."""
    val = map.pop(key)
    return _RemoveOut(map, val)


async def _sequence(_client, _stack, inputs: StructValue) -> StructValue:
    invals = cast(dict[str, IncomingWireType], inputs.values)
    first = deepcopy(cast(GraphValue, invals.pop("first")).value)
    second = deepcopy(cast(GraphValue, invals.pop("second")).value)
    newg = TierkreisGraph()
    outs1 = newg.insert_graph(
        first, **{port: newg.input[port] for port in first.inputs()}
    )
    outs2 = newg.insert_graph(second, **outs1)
    newg.set_outputs(**outs2)
    return StructValue({"sequenced": GraphValue(newg)})


namespace.functions["sequence"] = Function(
    run=_sequence,
    declaration=FunctionDeclaration(
        type_scheme=TypeScheme(
            variables=[
                TypeSchemeVar(name="in", kind=Kind(row=Empty())),
                TypeSchemeVar(name="middle", kind=Kind(row=Empty())),
                TypeSchemeVar(name="out", kind=Kind(row=Empty())),
            ],
            body=Type(
                graph=GraphType(
                    inputs=RowType(
                        content={
                            "second": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="middle"),
                                    outputs=RowType(rest="out"),
                                )
                            ),
                            "first": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="in"),
                                    outputs=RowType(rest="middle"),
                                )
                            ),
                        }
                    ),
                    outputs=RowType(
                        content={
                            "sequenced": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="in"),
                                    outputs=RowType(rest="out"),
                                )
                            )
                        }
                    ),
                )
            ),
        ),
        description="Sequence two graphs, where the"
        " outputs of the first graph match the inputs of the second.",
        input_order=["first", "second"],
        output_order=["sequenced"],
    ),
)


async def _parallel(_client, _stack, inputs: StructValue) -> StructValue:
    invals = inputs.values
    left = deepcopy(cast(GraphValue, invals.pop("left")).value)
    right = deepcopy(cast(GraphValue, invals.pop("right")).value)
    newg = TierkreisGraph()
    outs1 = newg.insert_graph(
        left, **{port: newg.input[port] for port in left.inputs()}
    )
    outs2 = newg.insert_graph(
        right, **{port: newg.input[port] for port in right.inputs()}
    )
    newg.set_outputs(**outs1, **outs2)
    return StructValue({"value": GraphValue(newg)})


namespace.functions["parallel"] = Function(
    run=_parallel,
    declaration=FunctionDeclaration(
        type_scheme=TypeScheme(
            variables=[
                TypeSchemeVar(name="left_in", kind=Kind(row=Empty())),
                TypeSchemeVar(name="left_out", kind=Kind(row=Empty())),
                TypeSchemeVar(name="right_in", kind=Kind(row=Empty())),
                TypeSchemeVar(name="right_out", kind=Kind(row=Empty())),
                TypeSchemeVar(name="value_in", kind=Kind(row=Empty())),
                TypeSchemeVar(name="value_out", kind=Kind(row=Empty())),
            ],
            constraints=[
                Constraint(
                    partition=PartitionConstraint(
                        left=Type(var="left_in"),
                        right=Type(var="right_in"),
                        union=Type(var="value_in"),
                    )
                ),
                Constraint(
                    partition=PartitionConstraint(
                        left=Type(var="left_out"),
                        right=Type(var="right_out"),
                        union=Type(var="value_out"),
                    )
                ),
            ],
            body=Type(
                graph=GraphType(
                    inputs=RowType(
                        content={
                            "left": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="left_in"),
                                    outputs=RowType(rest="left_out"),
                                )
                            ),
                            "right": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="right_in"),
                                    outputs=RowType(rest="right_out"),
                                )
                            ),
                        }
                    ),
                    outputs=RowType(
                        content={
                            "value": Type(
                                graph=GraphType(
                                    inputs=RowType(rest="value_in"),
                                    outputs=RowType(rest="value_out"),
                                )
                            )
                        }
                    ),
                )
            ),
        ),
        description="Merges two input `Graph`s into a single output `Graph` "
        "with the disjoint union of their inputs and similarly their outputs.",
        input_order=["left", "right"],
        output_order=["value"],
    ),
)


@namespace.function(type_vars={"a": StarKind()})
async def switch(pred: bool, if_true: a, if_false: a) -> a:
    """Passes one or other of two inputs through according to a \
third, boolean, input."""
    return if_true if pred else if_false


@dataclass
class _UnpackPairOut(Generic[a, b], UnpackRow):
    first: a
    second: b


@namespace.function(type_vars={"a": StarKind(), "b": StarKind()})
async def unpack_pair(pair: TierkreisPair[a, b]) -> _UnpackPairOut[a, b]:
    "Splits an input `Pair` into two separate outputs."
    return _UnpackPairOut(pair.first, pair.second)


async def _unpack_struct(_client, _stack, ins: StructValue) -> StructValue:
    return cast(StructValue, ins.values["struct"])


namespace.functions["unpack_struct"] = Function(
    run=_unpack_struct,
    declaration=FunctionDeclaration(
        type_scheme=TypeScheme(
            variables=[TypeSchemeVar(name="fields", kind=Kind(row=Empty()))],
            body=Type(
                graph=GraphType(
                    inputs=RowType(
                        content={
                            "struct": Type(
                                struct=StructType(shape=RowType(rest="fields"))
                            )
                        }
                    ),
                    outputs=RowType(rest="fields"),
                )
            ),
        ),
        description="Destructure a single `Struct` input, outputting one value "
        "per field in the input, output ports matching field names.",
        input_order=["struct"],
    ),
)


@namespace.function()
async def xor(a: bool, b: bool) -> bool:
    """Check either a or b are true; a ^ b"""
    return a ^ b


@dataclass
class GraphIn(Generic[a], UnpackRow):
    value: a


@dataclass
class GraphOut(Generic[b], UnpackRow):
    value: b


@namespace.function(name="map", type_vars={"a": StarKind(), "b": StarKind()})
async def _map(thunk: RuntimeGraph[GraphIn[a], GraphOut[b]], value: list[a]) -> list[b]:
    """Runs a Graph input on each element of a `Vec` input and \
collects the results into a `Vec` output."""
    # We avoid the use of callbacks in the python builtins and so must
    # define map in the pyruntime
    raise NotImplementedError()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha1/controller/runtime_storage_api.proto, v1alpha1/controller/tierkreis.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class Graph(betterproto.Message):
    id: str = betterproto.string_field(1)
    definition: bytes = betterproto.bytes_field(101)


@dataclass(eq=False, repr=False)
class NodeId(betterproto.Message):
    prefix: List[str] = betterproto.string_field(1)
    node_index: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class NodeStatus(betterproto.Message):
    id: "NodeId" = betterproto.message_field(1)
    job_id: str = betterproto.string_field(2)
    expected_duration_sec: Optional[int] = betterproto.uint32_field(3, optional=True)
    retry_count: int = betterproto.uint32_field(100)
    finished_at: Optional[datetime] = betterproto.message_field(201, optional=True)


@dataclass(eq=False, repr=False)
class RecordOutputRequest(betterproto.Message):
    label: str = betterproto.string_field(1)
    value: bytes = betterproto.bytes_field(2)
    job_id: str = betterproto.string_field(101)
    """Relationship fields."""


@dataclass(eq=False, repr=False)
class RecordOutputResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RecordJobFinishedRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    error_message: Optional[str] = betterproto.string_field(2, optional=True)


@dataclass(eq=False, repr=False)
class RecordJobFinishedResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RecordNodeRunRequest(betterproto.Message):
    id: "NodeId" = betterproto.message_field(1)
    attempt_id: int = betterproto.uint32_field(2)
    expected_duration_sec: Optional[int] = betterproto.uint32_field(3, optional=True)
    job_id: str = betterproto.string_field(101)
    """Relationship fields."""


@dataclass(eq=False, repr=False)
class RecordNodeRunResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RecordNodeFinishedRequest(betterproto.Message):
    id: "NodeId" = betterproto.message_field(1)
    attempt_id: int = betterproto.uint32_field(2)
    outputs: bytes = betterproto.bytes_field(3)
    job_id: str = betterproto.string_field(101)
    """Relationship fields."""


@dataclass(eq=False, repr=False)
class RecordNodeFinishedResponse(betterproto.Message):
    pass


class CheckpointRecordingServiceStub(betterproto.ServiceStub):
    async def record_output(
        self,
        record_output_request: "RecordOutputRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RecordOutputResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.controller.CheckpointRecordingService/RecordOutput",
            record_output_request,
            RecordOutputResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def record_job_finished(
        self,
        record_job_finished_request: "RecordJobFinishedRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RecordJobFinishedResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.controller.CheckpointRecordingService/RecordJobFinished",
            record_job_finished_request,
            RecordJobFinishedResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def record_node_run(
        self,
        record_node_run_request: "RecordNodeRunRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RecordNodeRunResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.controller.CheckpointRecordingService/RecordNodeRun",
            record_node_run_request,
            RecordNodeRunResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def record_node_finished(
        self,
        record_node_finished_request: "RecordNodeFinishedRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RecordNodeFinishedResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.controller.CheckpointRecordingService/RecordNodeFinished",
            record_node_finished_request,
            RecordNodeFinishedResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class CheckpointRecordingServiceBase(ServiceBase):

    async def record_output(
        self, record_output_request: "RecordOutputRequest"
    ) -> "RecordOutputResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def record_job_finished(
        self, record_job_finished_request: "RecordJobFinishedRequest"
    ) -> "RecordJobFinishedResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def record_node_run(
        self, record_node_run_request: "RecordNodeRunRequest"
    ) -> "RecordNodeRunResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def record_node_finished(
        self, record_node_finished_request: "RecordNodeFinishedRequest"
    ) -> "RecordNodeFinishedResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_record_output(
        self, stream: "grpclib.server.Stream[RecordOutputRequest, RecordOutputResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.record_output(request)
        await stream.send_message(response)

    async def __rpc_record_job_finished(
        self,
        stream: "grpclib.server.Stream[RecordJobFinishedRequest, RecordJobFinishedResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.record_job_finished(request)
        await stream.send_message(response)

    async def __rpc_record_node_run(
        self,
        stream: "grpclib.server.Stream[RecordNodeRunRequest, RecordNodeRunResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.record_node_run(request)
        await stream.send_message(response)

    async def __rpc_record_node_finished(
        self,
        stream: "grpclib.server.Stream[RecordNodeFinishedRequest, RecordNodeFinishedResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.record_node_finished(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha1.controller.CheckpointRecordingService/RecordOutput": grpclib.const.Handler(
                self.__rpc_record_output,
                grpclib.const.Cardinality.UNARY_UNARY,
                RecordOutputRequest,
                RecordOutputResponse,
            ),
            "/tierkreis.v1alpha1.controller.CheckpointRecordingService/RecordJobFinished": grpclib.const.Handler(
                self.__rpc_record_job_finished,
                grpclib.const.Cardinality.UNARY_UNARY,
                RecordJobFinishedRequest,
                RecordJobFinishedResponse,
            ),
            "/tierkreis.v1alpha1.controller.CheckpointRecordingService/RecordNodeRun": grpclib.const.Handler(
                self.__rpc_record_node_run,
                grpclib.const.Cardinality.UNARY_UNARY,
                RecordNodeRunRequest,
                RecordNodeRunResponse,
            ),
            "/tierkreis.v1alpha1.controller.CheckpointRecordingService/RecordNodeFinished": grpclib.const.Handler(
                self.__rpc_record_node_finished,
                grpclib.const.Cardinality.UNARY_UNARY,
                RecordNodeFinishedRequest,
                RecordNodeFinishedResponse,
            ),
        }



"""Utilities for building tierkreis graphs."""

import copy
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import betterproto
import networkx as nx

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
from tierkreis.core.function import FunctionName
from tierkreis.core.types import TierkreisType
from tierkreis.core.values import T, TierkreisValue

if TYPE_CHECKING:
    from tierkreis.builder import Unpack, ValueSource

PortID = str
Location = pg.Location


class TierkreisNode(ABC):
    """A node in a Tierkreis graph."""

    @abstractmethod
    def to_proto(self) -> pg.Node:
        """Convert to node protobuf message."""
        pass

    @classmethod
    def from_proto(cls, node: pg.Node) -> "TierkreisNode":
        """Load from protobuf node."""
        name, out_node = betterproto.which_one_of(node, "node")

        result: TierkreisNode
        if name == "const":
            result = ConstNode(TierkreisValue.from_proto(cast(pg.Value, out_node)))
        elif name == "box":
            box_node = cast(pg.BoxNode, out_node)
            result = BoxNode(
                graph=TierkreisGraph.from_proto(box_node.graph), location=box_node.loc
            )
        elif name == "function":
            fn_node = cast(pg.FunctionNode, out_node)
            result = FunctionNode(
                FunctionName.from_proto(fn_node.name), fn_node.retry_secs
            )
        elif name == "input":
            result = InputNode()
        elif name == "output":
            result = OutputNode()
        elif name == "tag":
            result = TagNode(cast(str, out_node))
        elif name == "match":
            result = MatchNode()
        else:
            raise ValueError(f"Unknown protobuf node type: {name}")

        if not isinstance(result, cls):
            raise TypeError()

        return result

    def is_discard_node(self) -> bool:
        """Delete nodes have some special behaviour, check for it."""
        return False

    def is_copy_node(self) -> bool:
        """Whether this node is a copy function node."""
        return False

    def is_unpack_node(self) -> bool:
        """Whether this node is an unpack struct function node."""
        return False


@dataclass(frozen=True)
class InputNode(TierkreisNode):
    """Graph input node."""

    def to_proto(self) -> pg.Node:
        return pg.Node(input=pg.Empty())


@dataclass(frozen=True)
class OutputNode(TierkreisNode):
    """Graph output node."""

    def to_proto(self) -> pg.Node:
        return pg.Node(output=pg.Empty())


@dataclass(frozen=True)
class ConstNode(TierkreisNode):
    """A node that holds a constant value."""

    value: TierkreisValue

    def to_proto(self) -> pg.Node:
        return pg.Node(const=self.value.to_proto())


@dataclass(frozen=True)
class BoxNode(TierkreisNode):
    """Nested subgraph node. Node signature is that of inner graph."""

    graph: "TierkreisGraph"
    location: Location

    def to_proto(self) -> pg.Node:
        return pg.Node(box=pg.BoxNode(loc=self.location, graph=self.graph.to_proto()))


@dataclass(frozen=True)
class FunctionNode(TierkreisNode):
    """A function call. Optional `retry_secs` specifies how long to wait before
    retrying.
    """

    function_name: FunctionName
    retry_secs: Optional[int] = None  # Stored as uint32, so always >=0

    def to_proto(self) -> pg.Node:
        return pg.Node(
            function=pg.FunctionNode(
                name=self.function_name.to_proto(), retry_secs=self.retry_secs
            )
        )

    def is_discard_node(self) -> bool:
        """Whether this node is a discard function node."""
        return self.function_name == FunctionName("discard")

    def is_copy_node(self) -> bool:
        return self.function_name == FunctionName("copy")

    def is_unpack_node(self) -> bool:
        return self.function_name == FunctionName("unpack_struct")


@dataclass(frozen=True)
class TagNode(TierkreisNode):
    """Node to tag a value and return a tagged variant."""

    tag_name: str

    def to_proto(self) -> pg.Node:
        return pg.Node(tag=self.tag_name)


@dataclass(frozen=True)
class MatchNode(TierkreisNode):
    """Match node. Returns the correct handler graph based on the incoming variant value."""

    def to_proto(self) -> pg.Node:
        return pg.Node(match=pg.Empty())


@dataclass(frozen=True)
class NodeRef(Iterable):
    """A reference to a node in a Tierkreis graph."""

    idx: int
    graph: "TierkreisGraph"
    outports: Optional[list[str]] = field(default=None, compare=False)

    def with_retry_secs(self, retry_secs: int) -> "NodeRef":
        """If this NodeRef refers to a FunctionNode, set its retry time in seconds.
        Otherwise, raises ValueError.
        """
        node = self.graph[self.idx]
        if not isinstance(node, FunctionNode):
            raise ValueError(
                "{self} must refer to a FunctionNode, but actually refers to {node}"
            )
        self.graph[self.idx] = FunctionNode(node.function_name, retry_secs)
        return self

    def default_nodeport(self, create: bool = True) -> "NodePort":
        """Output port name to use if none is specified.

        Raises:
            ValueError: If there are multiple known output ports.

        Returns:
            Single output port if exactly one is known, else "value".
        """
        if self.outports:
            if len(self.outports) != 1:
                raise ValueError(f"Cannot choose one outport from {self.outports}")
            return self[self.outports[0]]
        return self["value"]

    def __getitem__(self, key: str) -> "NodePort":
        # syntactic sugar for creating references to output ports from node
        return NodePort(self, key)

    def __len__(self) -> int:
        return len(self.outports or [])

    def __iter__(self) -> Iterator["NodePort"]:
        return (self[p] for p in (self.outports or []))


@dataclass(frozen=True)
class NodePort:
    """A named port of a node in a Tierkreis graph."""

    node_ref: NodeRef
    port: PortID

    def __getitem__(self, field_name: str) -> "Unpack":
        raise NotImplementedError(
            "Import tierkreis.builder to get"
            "automatic unpacking of structs using the [<field>] syntax. "
        )


@dataclass(frozen=True)
class TierkreisEdge:
    """An edge in a Tierkreis graph, from output `NodePort` to input `NodePort`.
    Optionally records a type annotation for the value carried by the edge.
    """

    source: NodePort
    target: NodePort
    type_: Optional[TierkreisType]

    def to_proto(self) -> pg.Edge:
        edge_type = None if self.type_ is None else self.type_.to_proto()
        return pg.Edge(
            port_from=self.source.port,
            node_from=self.source.node_ref.idx,
            port_to=self.target.port,
            node_to=self.target.node_ref.idx,
            edge_type=cast(pg.Type, edge_type),
        )

    def to_edge_handle(self) -> "_EdgeData":
        return (
            self.source.node_ref.idx,
            self.target.node_ref.idx,
            (self.source.port, self.target.port),
        )


# allow specifying input as a port, node with single output
IncomingWireType = Union[NodePort, NodeRef]


def to_nodeport(n: IncomingWireType) -> NodePort:
    """Convert an incoming wire type to a NodePort. Assumes default output port
    for a `NodeRef`."""
    return n if isinstance(n, NodePort) else n.default_nodeport()


_EdgeData = Tuple[int, int, Tuple[str, str]]


def _to_edgedata(edgeit: Iterable[Any]) -> Iterator[_EdgeData]:
    return (
        (src, dest, (cast(str, srcport), cast(str, destport)))
        for (src, dest, (srcport, destport)) in edgeit
    )


class MismatchedGraphs(Exception):
    def __init__(self, graph: "TierkreisGraph", endpoint: NodeRef):
        self.graph = graph
        self.endpoint = endpoint
        super().__init__("Cannot add an edge whose endpoint is on a different graph")


class TierkreisGraph:
    """Python in-memory representation of a Tierkreis graph. Can be loaded from
    and written to protobuf format. Supports building graphs programmatically.
    """

    input_node_idx: int = 0
    output_node_idx: int = 1

    @dataclass(frozen=True)
    class MissingEdge(Exception):
        source: NodePort
        target: NodePort

    def __init__(self, name: str = "") -> None:
        """Create an empty graph with an optional name."""
        self.name = name
        self._graph = nx.MultiDiGraph()
        inp = self.add_node(InputNode())
        assert inp.idx == self.input_node_idx
        output = self.add_node(OutputNode())
        assert output.idx == self.output_node_idx

        self.input_order: list[str] = []
        self.output_order: list[str] = []

    @property
    def input(self) -> NodeRef:
        """Get a reference to the input node of the graph."""
        return NodeRef(self.input_node_idx, self)

    @property
    def output(self) -> NodeRef:
        """Get a reference to the output node of the graph."""
        return NodeRef(self.output_node_idx, self)

    def inputs(self) -> List[str]:
        """A list of all output ports of the input node (the inputs to the graph)."""
        return [edge.source.port for edge in self.out_edges(self.input)]

    def outputs(self) -> List[str]:
        """A list of all input ports of the output node (the outputs of the graph)."""
        return [edge.target.port for edge in self.in_edges(self.output)]

    @property
    def n_nodes(self) -> int:
        """The number of nodes in the graph."""
        return len(self._graph)

    def add_node(
        self,
        _tk_node: TierkreisNode,
        /,
        **incoming_wires: IncomingWireType,
    ) -> NodeRef:
        """Add a node to the graph, connecting it to the specified inputs.
        The node to add is the only positional argument, the remaining keyword
        arguments specify the input port and the sources to connect to those ports.

        Args:
            _tk_node: The node to add to the graph.

        Returns:
            NodeRef: A reference to the added node.
        """
        node_ref = NodeRef(self._graph.number_of_nodes(), self)
        self._graph.add_node(node_ref.idx, node_info=_tk_node)
        for target_port_name, source in incoming_wires.items():
            self.add_edge(source, node_ref[target_port_name])

        return node_ref

    def add_func(
        self,
        _tk_function: Union[str, FunctionName],
        _retry_secs: Optional[int] = None,
        /,
        **kwargs: IncomingWireType,
    ) -> NodeRef:
        """Add a function call to the graph, connecting it to the specified inputs.

        Args:
            _tk_function: Name of the
                function to call.
            _retry_secs: Optionally specify
                how long to wait before retrying, defaults to None

        Returns:
            NodeRef: Reference to the added node.
        """
        f_name: FunctionName = (
            FunctionName.parse(_tk_function)
            if isinstance(_tk_function, str)
            else _tk_function
        )

        return self.add_node(FunctionNode(f_name, _retry_secs), **kwargs)

    def add_const(self, value: Any) -> NodeRef:
        """Add a constant value to the graph. Non-tierkreis values are converted
        if possible.
        """
        return self.add_node(ConstNode(TierkreisValue.from_python(value)))

    def add_box(
        self,
        graph: "TierkreisGraph",
        /,
        **kwargs: IncomingWireType,
    ) -> NodeRef:
        """Add a box node by specifying the inner graph and the inputs to
        connect to the node.
        """
        # TODO restrict to graph i/o
        return self.add_node(BoxNode(graph, Location([])), **kwargs)

    def add_match(
        self,
        variant_value: IncomingWireType,
        /,
        **variant_handlers: IncomingWireType,
    ) -> NodeRef:
        """Add a `Match` node to the graph. The variant value is the only
        positional argument, handlers are specified as keyword arguments.
        """
        return self.add_node(
            MatchNode(), variant_value=variant_value, **variant_handlers
        )

    def add_tag(
        self,
        tag: str,
        *,  # Following are keyword only
        value: IncomingWireType,
    ) -> NodeRef:
        """Add a tag node to the graph. Tag the value with the specified tag string."""
        return self.add_node(TagNode(tag), value=value)

    def insert_graph(
        self,
        graph: "TierkreisGraph",
        /,
        **kwargs: IncomingWireType,
    ) -> Dict[str, IncomingWireType]:
        """Given a graph and wires to connect to that graph's inputs,
        inline that graph into <self>, and return the outputs of
        the inserted graph.
        """

        index_map: Dict[int, int] = dict()
        for node_idx, node in enumerate(graph.nodes()):
            if node_idx not in {graph.input_node_idx, graph.output_node_idx}:
                index_map[node_idx] = self.add_node(node).idx

        return_outputs: Dict[str, IncomingWireType] = {}

        for edge in graph.edges():
            source_node = edge.source.node_ref.idx
            target_node = edge.target.node_ref.idx
            source_port = (
                kwargs[edge.source.port]
                if source_node == graph.input_node_idx
                else NodeRef(index_map[source_node], self)[edge.source.port]
            )

            if target_node == graph.output_node_idx:
                return_outputs[edge.target.port] = source_port
            else:
                target_port = NodeRef(index_map[target_node], self)[edge.target.port]

                self.add_edge(source_port, target_port, edge.type_)

        return return_outputs

    def inline_boxes(self, recursive=False) -> "TierkreisGraph":
        """Inline boxes by inserting the graphs they contain in to the parent
        graph. Optionally do this recursively.

        Returns:
            TierkreisGraph: Inlined graph
        """

        graph = copy.deepcopy(self)
        deleted_nodes: list[int] = []
        # Iterate through self.nodes rather than graph.nodes so we can mutate latter
        for node_idx, node in enumerate(self.nodes()):
            if not isinstance(node, BoxNode):
                continue

            boxed_g = node.graph
            if recursive:
                boxed_g = boxed_g.inline_boxes(recursive)

            curr_inputs = graph.in_edges(node_idx)

            incoming_ports = {edge.target.port: edge.source for edge in curr_inputs}
            curr_outputs = [
                graph._to_tkedge(e)
                for e in _to_edgedata(graph._graph.out_edges(node_idx, keys=True))
            ]
            # Removal all incident edges
            graph._graph.remove_edges_from(
                list(graph._graph.out_edges(node_idx))
                + list(graph._graph.in_edges(node_idx))
            )

            deleted_nodes.append(node_idx)

            inserted_outputs = graph.insert_graph(boxed_g, **incoming_ports)

            for edge in curr_outputs:
                graph.add_edge(
                    inserted_outputs[edge.source.port], edge.target, edge.type_
                )
        if deleted_nodes:
            graph.remove_nodes(deleted_nodes)

        return graph

    def nodes(self) -> Iterator[TierkreisNode]:
        """Iterator over all nodes in the graph."""
        return (
            self._graph.nodes[idx]["node_info"]
            for idx in range(self._graph.number_of_nodes())
        )

    def edges(self) -> Iterator[TierkreisEdge]:
        """Iterator over all edges in the graph."""
        return map(self._to_tkedge, _to_edgedata(self._graph.edges(keys=True)))

    def __getitem__(self, key: Union[int, NodeRef]) -> TierkreisNode:
        name = key.idx if isinstance(key, NodeRef) else key
        return self._graph.nodes[name]["node_info"]

    def __setitem__(self, key: Union[int, NodeRef], node: TierkreisNode):
        name = key.idx if isinstance(key, NodeRef) else key
        self._graph.nodes[name]["node_info"] = node

    def add_edge(
        self,
        source: IncomingWireType,
        node_port_to: NodePort,
        edge_type: Type | TierkreisType | None = None,
    ) -> TierkreisEdge:
        """Add an edge to the graph, connecting output port to input port.

        Args:
            source: Source of the edge, either an
                output port or a node with a single output.
            node_port_to: Target input port.
            edge_type: Optional type annotation, defaults to None

        Raises:
            MismatchedGraphs: If source and target nodes are not in the
                same graph.
            ValueError: If there is already an edge from source to
                target.

        Returns:
            TierkreisEdge: Reference to the added edge.
        """
        tk_type = _to_tierkreis_type(edge_type)
        node_port_from = to_nodeport(source)
        if node_port_from.node_ref.graph is not self:
            raise MismatchedGraphs(self, node_port_from.node_ref)
        if node_port_to.node_ref.graph is not self:
            raise MismatchedGraphs(self, node_port_to.node_ref)

        # if port is currently connected to discard, replace that edge
        existing_edge = self.out_edge_from_port(node_port_from)
        if isinstance(existing_edge, TierkreisType):
            tk_type = existing_edge
        elif existing_edge is not None:
            raise ValueError(
                f"Already an edge from {node_port_from} to {existing_edge.target}"
            )
        edge_data = (
            node_port_from.node_ref.idx,
            node_port_to.node_ref.idx,
            (node_port_from.port, node_port_to.port),
        )
        self._graph.add_edge(
            *edge_data,
            type=tk_type,
        )

        return self._to_tkedge(edge_data)

    def remove_edge(self, edge: TierkreisEdge):
        """Remove an edge from the graph."""
        assert edge.source.node_ref.graph is self
        assert edge.target.node_ref.graph is self
        self._graph.remove_edge(
            edge.source.node_ref.idx,
            edge.target.node_ref.idx,
            (edge.source.port, edge.target.port),
        )

    def remove_nodes(self, nodes: Iterable[int]):
        """Remove nodes from the graph by index."""
        self._graph.remove_nodes_from(nodes)
        # shift node indices to account for missing nodes
        # without this step indices will not be contiguous
        nx.relabel_nodes(
            self._graph,
            {n: i for i, n in enumerate(sorted(self._graph.nodes()))},
            copy=False,
        )

    def annotate_input(
        self, input_port: str, edge_type: Optional[Union[Type, TierkreisType]]
    ):
        """Annotate an input port of the graph with a type."""
        (in_edge,) = [
            e
            for e in _to_edgedata(self._graph.out_edges(self.input_node_idx, keys=True))
            if e[2][0] == input_port
        ]
        tk_type = _to_tierkreis_type(edge_type)
        self._graph.edges[in_edge]["type"] = tk_type

    def annotate_output(
        self, output_port: str, edge_type: Optional[Union[Type, TierkreisType]]
    ):
        """Annotate an output port of the graph with a type."""
        (out_edge,) = [
            e
            for e in _to_edgedata(self._graph.in_edges(self.output_node_idx, keys=True))
            if e[2][1] == output_port
        ]

        tk_type = _to_tierkreis_type(edge_type)
        self._graph.edges[out_edge]["type"] = tk_type

    def get_edge(self, source: NodePort, target: NodePort) -> TierkreisEdge:
        """Retrieve an edge from the graph by source and target ports.
        Raises `MissingEdge` if the edge does not exist.
        """
        try:
            return next(
                e for e in self.out_edges(source.node_ref) if e.target == target
            )
        except StopIteration as e:
            raise self.MissingEdge(source, target) from e

    def set_outputs(self, **kwargs: IncomingWireType) -> None:
        """Set outputs of the graph by keyword arguments mapping output port
        names to incoming wires.
        """
        for out_name, port in kwargs.items():
            self.add_edge(port, self.output[out_name])

    def _to_tkedge(self, handle: _EdgeData) -> TierkreisEdge:
        src, tgt, (src_port, tgt_port) = handle
        edge_data = cast(
            Mapping, self._graph.get_edge_data(src, tgt, (src_port, tgt_port))
        )
        return TierkreisEdge(
            NodeRef(src, self)[src_port],
            NodeRef(tgt, self)[tgt_port],
            edge_data["type"],
        )

    def in_edges(self, node: Union[NodeRef, int]) -> Iterator[TierkreisEdge]:
        """Iterator over incoming edges to a node."""
        node_name = node if isinstance(node, int) else node.idx
        return map(
            self._to_tkedge, _to_edgedata(self._graph.in_edges(node_name, keys=True))
        )

    def out_edges(self, node: Union[NodeRef, int]) -> Iterator[TierkreisEdge]:
        """Iterator over outgoing edges from a node."""
        node_idx = node if isinstance(node, int) else node.idx
        return map(
            self._to_tkedge,
            _to_edgedata(self._graph.out_edges(node_idx, keys=True)),
        )

    def discard(self, out_port: NodePort) -> None:
        """Add a discard node, discarding the value at the specified output port."""
        _ = self.add_func("discard", value=out_port)

    def make_pair(
        self, first_port: IncomingWireType, second_port: IncomingWireType
    ) -> NodePort:
        """Utility function to create a pair from two values."""
        make_n = self.add_func("make_pair", first=first_port, second=second_port)
        return make_n["pair"]

    def unpack_pair(self, pair_port: IncomingWireType) -> Tuple[NodePort, NodePort]:
        """Utility function to unpack a pair into two values."""
        unp_n = self.add_func("unpack_pair", pair=pair_port)
        return unp_n["first"], unp_n["second"]

    def make_vec(self, element_ports: Iterable[IncomingWireType]) -> NodePort:
        """Utility function to create a vector from an iterator of incoming wires."""
        vec_port = self.add_const(list())["value"]

        for port in element_ports:
            vec_port = self.add_func("push", vec=vec_port, item=port)["vec"]

        return vec_port

    def vec_last_n_elems(self, vec_port: NodePort, n_elements: int) -> List[NodePort]:
        """Utility to pop `n_elements` from a vector and return the
        corresponding output ports.
        """
        curr_arr = vec_port
        outports: List[NodePort] = []
        for _ in range(n_elements):
            pop = self.add_func("pop", vec=curr_arr)
            curr_arr = pop["vec"]
            outports.insert(0, pop["item"])
        return outports

    def copy_value(self, value: IncomingWireType) -> Tuple[NodePort, NodePort]:
        """Copy a value to two output ports."""
        copy_n = self.add_func("copy", value=value)

        return copy_n["value_0"], copy_n["value_1"]

    def out_edge_from_port(self, source: NodePort) -> Optional[TierkreisEdge]:
        """If there is an edge at port, return it, else None."""
        out_edges = [
            out_edge
            for out_edge in self.out_edges(source.node_ref)
            if out_edge.source.port == source.port
        ]
        if len(out_edges) == 0:
            return None
        (out_edge,) = out_edges
        return out_edge

    def to_proto(self) -> pg.Graph:
        """Build protobuf message from graph."""
        pg_graph = pg.Graph()
        pg_graph.nodes = [n.to_proto() for n in self.nodes()]
        pg_graph.edges = [e.to_proto() for e in self.edges()]
        pg_graph.name = self.name
        pg_graph.input_order = self.input_order
        pg_graph.output_order = self.output_order
        return pg_graph

    @classmethod
    def from_proto(cls, pg_graph: pg.Graph) -> "TierkreisGraph":
        """Load from protobuf message."""
        tk_graph = cls()
        tk_graph.name = pg_graph.name
        tk_graph.input_order = pg_graph.input_order
        tk_graph.output_order = pg_graph.output_order

        for pgn in pg_graph.nodes[2:]:
            tk_graph.add_node(TierkreisNode.from_proto(pgn))

        for pg_edge in pg_graph.edges:
            source_node = NodeRef(pg_edge.node_from, tk_graph)
            target_node = NodeRef(pg_edge.node_to, tk_graph)
            source = NodePort(source_node, PortID(pg_edge.port_from))
            target = NodePort(target_node, PortID(pg_edge.port_to))
            # TODO make all edge type conversions possible
            edge_type = pg_edge.edge_type
            tk_edge_type: Optional[TierkreisType] = None
            if edge_type is not None:
                try:
                    tk_edge_type = TierkreisType.from_proto(edge_type)
                except ValueError:
                    tk_edge_type = None

            tk_graph.add_edge(source, target, tk_edge_type)
        return tk_graph

    def to_python(self, type_: typing.Type[T]) -> T:
        if isinstance(type_, typing.TypeVar) or type_ is TierkreisGraph:
            return cast(T, self)
        raise TypeError()

    def __call__(
        self: "TierkreisGraph", *args: "ValueSource", **kwargs: "ValueSource"
    ) -> NodeRef:
        raise NotImplementedError(
            "TierkreisGraph can only be called inside builder contexts."
        )


def _to_tierkreis_type(
    edge_type: Optional[Union[Type, TierkreisType]],
) -> Optional[TierkreisType]:
    if edge_type is None:
        return None
    else:
        return (
            edge_type
            if isinstance(edge_type, TierkreisType)
            else TierkreisType.from_python(edge_type)
        )


# GraphValue defined after TierkreisGraph to avoid circular/delayed import


@dataclass(frozen=True)
class GraphValue(TierkreisValue):
    """A value defined by a Tierkreis graph. (Higher order value)."""

    value: TierkreisGraph
    _proto_name: str = "graph"

    @property
    def _instance_pytype(self) -> typing.Type:
        return TierkreisGraph

    def to_proto(self) -> pg.Value:
        return pg.Value(graph=self.value.to_proto())

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        from tierkreis.core.python import RuntimeGraph

        if isinstance(type_, typing.TypeVar):
            return cast(T, self)
        if typing.get_origin(type_) is RuntimeGraph:
            return cast(T, RuntimeGraph(self.value))
        if issubclass(type_, TierkreisGraph):
            return cast(T, self.value)

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        return cls(TierkreisGraph.from_proto(cast(pg.Graph, value)))

    def __str__(self) -> str:
        return "GraphValue"

    def viz_str(self) -> str:
        return "Graph"


# allow graph displays in jupyter notebooks
from tierkreis.core.graphviz import tierkreis_to_graphviz  # noqa: E402

setattr(
    TierkreisGraph,
    "_repr_svg_",
    lambda self: tierkreis_to_graphviz(self)._repr_svg_(),  # type: ignore
)



"""Simple pure-Python implementation of the Tierkreis runtime protocol. Allows execution of
graphs with Python workers. Does not support type checking or connecting to
workers over the network."""

from .python_runtime import PyRuntime



from dataclasses import dataclass
from typing import Generic, Type, TypeVar, cast

import pydantic as pyd
import pytest
from test_worker.main import (
    MyGeneric,
    MyGenericList,
)

from tierkreis.builder import Const, Output, graph
from tierkreis.client.runtime_client import RuntimeClient
from tierkreis.core._internal import generic_origin
from tierkreis.core.types import (
    IntType,
    Row,
    StructType,
    TierkreisType,
    VarType,
    VecType,
)
from tierkreis.core.values import (
    IncompatibleAnnotatedValue,
    IntValue,
    StructValue,
    TierkreisValue,
    VecValue,
)


@pytest.mark.asyncio
async def test_generic(bi, client: RuntimeClient) -> None:
    py_type = MyGeneric[int]
    py_val = MyGeneric(x=1)
    py_val_explicit_type = MyGeneric[int](x=1)
    tk_val = StructValue(
        {
            "x": IntValue(1),
        }
    )
    tk_ty = StructType(shape=Row({"x": IntType()}), name="MyGeneric[int]")

    assert TierkreisType.from_python(py_type) == tk_ty
    assert TierkreisValue.from_python(py_val, py_type) == tk_val
    with pytest.raises(IncompatibleAnnotatedValue, match="~G"):
        # value doesn't know it is instance of concrete type so fails.
        TierkreisValue.from_python(py_val)
    # whereas the pydantic-validated instantiation records its own concrete type
    assert TierkreisValue.from_python(py_val_explicit_type) == tk_val
    assert tk_val.to_python(py_type) == py_val

    on = bi["python_nodes"]

    @graph()
    def g() -> Output:
        return Output(on.generic_int_to_str(Const(py_val, py_type)))

    out = (await client.run_graph(g))["value"]

    assert out.to_python(MyGeneric[str]) == MyGeneric(x="got number: 1")


@pytest.mark.asyncio
async def test_generic_list(bi, client: RuntimeClient) -> None:
    py_type = MyGenericList[int]
    py_val = MyGenericList(x=[1])
    py_val_explicit_type = MyGenericList[int](x=[1])
    tk_val = StructValue(
        {
            "x": VecValue([IntValue(1)]),
        }
    )
    tk_ty = StructType(shape=Row({"x": VecType(IntType())}), name="MyGenericList[int]")

    assert TierkreisType.from_python(py_type) == tk_ty
    assert TierkreisValue.from_python(py_val, py_type) == tk_val
    with pytest.raises(IncompatibleAnnotatedValue, match="~G"):
        # value doesn't know it is instance of concrete type so fails.
        TierkreisValue.from_python(py_val)
    # whereas the pydantic-validated instantiation records its own concrete type
    assert TierkreisValue.from_python(py_val_explicit_type) == tk_val
    assert tk_val.to_python(py_type) == py_val

    on = bi["python_nodes"]

    @graph()
    def g() -> Output:
        return Output(on.generic_int_to_str_list(Const(py_val, py_type)))

    out = (await client.run_graph(g))["value"]

    assert out.to_python(MyGenericList[str]) == MyGenericList(x=["got number: 1"])


@pytest.mark.xfail(reason="generic dataclasses aren't supported - only pydantic models")
def test_generic_dataclass() -> None:
    from typing import Generic, TypeVar

    T = TypeVar("T")

    @dataclass
    class MyGenericDataclass(Generic[T]):
        x: T

    assert TierkreisType.from_python(MyGenericDataclass[int]) == StructType(
        shape=Row({"x": VarType("T")}), name="MyGenericDataclass"
    )
    tk_val = StructValue(
        {
            "x": IntValue(1),
        }
    )
    py_val = MyGenericDataclass(x=1)
    # both asserts below fail
    assert tk_val.to_python(MyGenericDataclass[int]) == py_val
    assert TierkreisValue.from_python(py_val, MyGenericDataclass[int]) == tk_val


def test_pydantic_nested() -> None:
    T = TypeVar("T")

    class Inner(pyd.BaseModel, Generic[T]):
        items: list[T]  # Something type-erased

    class Outer(pyd.BaseModel, Generic[T]):
        val: Inner[T]

    # All these pass type checking (consistent with type erasure):
    # (Note: make_inner_ab: a,b = presence of annotation on constructor, return-type)
    def make_inner_validated() -> Inner:
        return Inner[int](items=[3, 4, 5])

    def make_inner() -> Inner[int]:
        return Inner(items=[3, 5, 7])

    # Pydantic is happy with all these at runtime:
    un_un: Outer = Outer(val=make_inner())
    val_un: Outer = Outer(val=make_inner_validated())
    val_val: Outer = Outer[int](val=make_inner_validated())

    # pydantic is happy with all of these at runtime;
    # Tierkreis can convert only the validated without an explicit type annotation
    for o in [un_un, val_un]:
        with pytest.raises(IncompatibleAnnotatedValue):
            TierkreisValue.from_python(o)
    tk_val = TierkreisValue.from_python(val_val)
    assert tk_val.to_python(Outer[int]) == val_val  # and back (with explicit type)

    # Tierkreis can convert any of them with an explicit type annotation:
    for o in [un_un, val_un, val_val]:
        tk_val = TierkreisValue.from_python(o, Outer[int])
        assert tk_val.to_python(Outer[int]) == o

    # And pydantic rejects this one at runtime:
    with pytest.raises(
        pyd.ValidationError,
        match=r"Input should be a valid dictionary or instance of Inner\[int\]",
    ):
        Outer[int](val=make_inner())


def test_generic_origin() -> None:
    assert generic_origin(list[int]) is list
    assert generic_origin(list) is None

    @dataclass
    class MyDataclass:
        foo: int

    assert generic_origin(MyDataclass) is None

    T = TypeVar("T")

    @dataclass
    class MyGenericDataclass(Generic[T]):
        item: T

    assert generic_origin(MyGenericDataclass[int]) == MyGenericDataclass
    assert generic_origin(MyGenericDataclass) is None

    class MyModel(pyd.BaseModel):
        f: int

    assert generic_origin(MyModel) is None

    class MyGenericModel(pyd.BaseModel, Generic[T]):
        items: list[T]

    assert generic_origin(MyGenericModel[int]) == MyGenericModel
    assert generic_origin(MyGenericModel) is None

    assert generic_origin(cast(Type, T)) is None



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha1/graph.proto
# plugin: python-betterproto
# This file has been @generated
import builtins
from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)

import betterproto


@dataclass(eq=False, repr=False)
class Location(betterproto.Message):
    """Structured name identifying a path from a root to a Runtime"""

    location: List[str] = betterproto.string_field(1)
    """
    Atoms in the name, each identifying a Runtime that is a child
     of the previous Runtime
    """


@dataclass(eq=False, repr=False)
class StructValue(betterproto.Message):
    """A `Value::struct` of unordered fields named by distinct strings"""

    map: Dict[str, "Value"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """Fields keyed by name"""


@dataclass(eq=False, repr=False)
class PairValue(betterproto.Message):
    """
    Contents of a `Value::pair` or one entry in a `MapValue`, i.e. two values
    """

    first: "Value" = betterproto.message_field(1)
    """First element of the `Value::pair` or key of the map entry"""

    second: "Value" = betterproto.message_field(2)
    """Second element of the `Value::pair` or value/RHS of the map entry"""


@dataclass(eq=False, repr=False)
class MapValue(betterproto.Message):
    """Contents of a `Value::map`: a collection of (key, value) pairs"""

    pairs: List["PairValue"] = betterproto.message_field(1)
    """
    List of mappings.
     (Note: keys of a protobuf `map` cannot be message types like `Value`.)
    """


@dataclass(eq=False, repr=False)
class VecValue(betterproto.Message):
    """Contents of a `Value::vec`: a list of values"""

    vec: List["Value"] = betterproto.message_field(2)
    """Elements of the list"""


@dataclass(eq=False, repr=False)
class VariantValue(betterproto.Message):
    """
    A `Value` tagged with a string to make a disjoint union of component types
    """

    tag: str = betterproto.string_field(1)
    """
    Label, selects the appropriate handler of a `Node::match`
     destructuring the variant
    """

    value: "Value" = betterproto.message_field(2)
    """Contents"""


@dataclass(eq=False, repr=False)
class Value(betterproto.Message):
    """A value that can be passed around a Tierkreis graph."""

    graph: "Graph" = betterproto.message_field(1, group="value")
    """A Tierkreis Graph (i.e., a higher-order function value)"""

    integer: int = betterproto.int64_field(2, group="value")
    """Signed 64-bit integer value"""

    boolean: bool = betterproto.bool_field(3, group="value")
    """Boolean value"""

    str: builtins.str = betterproto.string_field(9, group="value")
    """String value"""

    flt: float = betterproto.double_field(10, group="value")
    """Double-precision (64-bit) floating-point value"""

    pair: "PairValue" = betterproto.message_field(4, group="value")
    """A pair of two other values"""

    vec: "VecValue" = betterproto.message_field(5, group="value")
    """An arbitrary-length list of values"""

    map: "MapValue" = betterproto.message_field(7, group="value")
    """
    A map from keys (values) to values. Keys must be hashable, i.e. must not
     be or contain `graph`s, `map`s, `struct`s or `float`s
    """

    struct: "StructValue" = betterproto.message_field(8, group="value")
    """
    A structure (aka record) value with string-named fields (each storing a `Value`)
    """

    variant: "VariantValue" = betterproto.message_field(12, group="value")
    """
    A value tagged with a string to make a value of sum type (see `Type::Variant`)
    """


@dataclass(eq=False, repr=False)
class Output(betterproto.Message):
    """
    A value was placed onto an edge - used for visualization by the Python runtime.
    """

    edge: "Edge" = betterproto.message_field(1)
    """The edge (of the outermost graph) onto which the value was placed"""

    value: "Value" = betterproto.message_field(2)
    """The value, i.e. an intermediate during graph execution"""


@dataclass(eq=False, repr=False)
class OutputStream(betterproto.Message):
    """
    A list of all the `Output`s, used for visualization by the Python runtime.
    """

    stream: List["Output"] = betterproto.message_field(1)
    """All the `Output`s so far."""


@dataclass(eq=False, repr=False)
class Type(betterproto.Message):
    """
    / A type of values (generally of `Kind::Star`) that can be passed around a Tierkreis graph
     (except `Row`, of `Kind::Row`, describing a row of values which is not a value itself).
    """

    var: str = betterproto.string_field(1, group="type")
    """
    Type variable, used inside polymorphic `TypeScheme`s only.
     Can be of [Kind::Row] or [Kind::Star].
    """

    int: "Empty" = betterproto.message_field(2, group="type")
    """Type of signed integers (values must be `Value::int`)"""

    bool: "Empty" = betterproto.message_field(3, group="type")
    """Type of booleans (values must be `Value::bool`)"""

    graph: "GraphType" = betterproto.message_field(4, group="type")
    """Type of `Value::graph`s, with a specified row of inputs and outputs"""

    pair: "PairType" = betterproto.message_field(5, group="type")
    """Type of `Value::pair`s, with types for the two elements"""

    vec: "Type" = betterproto.message_field(6, group="type")
    """
    Type of arbitrary-length homogeneous lists, with the type for all elements.
     Values must be `Value::vec`.
    """

    row: "RowType" = betterproto.message_field(7, group="type")
    """
    An unordered row of named types. Unlike the other variants (except possibly
     `Type::Var`), this is a `Kind::Row`, not a type of values (`Kind::Star`), so
     cannot be used as a member of any other Type (e.g. `Type::Vec`). However,
     this can appear in a `LacksConstraint` or `PartitionConstraint`, or in a
     `tierkreis.v1alpha.signature.UnifyError`.
    """

    map: "PairType" = betterproto.message_field(9, group="type")
    """
    Type of maps, containing two types - one for all the keys and one for the values.
     We do nothing to rule out key types that are not hashable, only the actual
     `Value`s used as keys.
    """

    struct: "StructType" = betterproto.message_field(10, group="type")
    """
    Type of `Value::struct`s with a specified row of field names and types.
     Optionally, the struct type itself may have a name.
    """

    str: "Empty" = betterproto.message_field(11, group="type")
    """Type of strings (values must be `Value::str`)"""

    flt: "Empty" = betterproto.message_field(12, group="type")
    """Type of double-precision floats ()"""

    variant: "RowType" = betterproto.message_field(14, group="type")
    """
    A disjoint (tagged) union of other types, given as a row.
     May be open, for the output of a `Tag` operation, or closed,
     for the input to match (where the handlers are known).
    """


@dataclass(eq=False, repr=False)
class GraphType(betterproto.Message):
    """
    Describes the type of a graph by specifying the input and output rows
    """

    inputs: "RowType" = betterproto.message_field(1)
    """The inputs to the graph (known and/or variable)"""

    outputs: "RowType" = betterproto.message_field(2)
    """The outputs to the graph (known and/or variable)"""


@dataclass(eq=False, repr=False)
class PairType(betterproto.Message):
    """Used to describe the type of a pair, or of a map (key + value)"""

    first: "Type" = betterproto.message_field(1)
    """Type of the first element of the pair, or the keys in the map"""

    second: "Type" = betterproto.message_field(2)
    """Type of the second element of the pair, or the values in the map"""


@dataclass(eq=False, repr=False)
class RowType(betterproto.Message):
    """
    An unordered set of names (strings), each with a type; possibly
     plus a variable (of `Kind::Row`) to stand for an unknown portion.
    """

    content: Dict[str, "Type"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """
    Known labels, and types for each (of course these may contain variables)
    """

    rest: str = betterproto.string_field(2)
    """
    May contain the name of a variable (of `Kind::Row`), in which case the `RowType`
     is an "open row" and also includes any number of other fields stood for by that
     variable (with names disjoint from those in `content`).
     Otherwise, a "closed row" i.e. exactly/only the fields in `content`
    """


@dataclass(eq=False, repr=False)
class StructType(betterproto.Message):
    """
    Describes a `Type::Struct`: the row of fields, optionally with a name for the type.
    """

    shape: "RowType" = betterproto.message_field(1)
    """The fields in the struct"""

    name: str = betterproto.string_field(2, group="name_opt")
    """Optional name of the struct type. (TODO how is this used?)"""


@dataclass(eq=False, repr=False)
class FunctionName(betterproto.Message):
    """Qualified name of a function"""

    namespaces: List[str] = betterproto.string_field(1)
    """
    Identifies a `tierkreis.v1alpha1.signature.Namespace` (perhaps a subspace)
     by traversing downwards from the root
    """

    name: str = betterproto.string_field(2)
    """Name of the function within that subspace"""


@dataclass(eq=False, repr=False)
class FunctionNode(betterproto.Message):
    """
    Describes a `Node::function`, identifying the function to call,
     and optionally a time after which execution should be retried
    """

    name: "FunctionName" = betterproto.message_field(1)
    """Qualified name of the function"""

    retry_secs: Optional[int] = betterproto.uint32_field(2, optional=True)
    """
    Time after which execution of the function may be assumed to have
     failed and should be retried (in due course)
    """


@dataclass(eq=False, repr=False)
class BoxNode(betterproto.Message):
    """Describes a `Node::box`: a location at which to run a graph"""

    loc: "Location" = betterproto.message_field(1)
    """On which runtime to run the subgraph"""

    graph: "Graph" = betterproto.message_field(2)
    """Definition of the graph to run"""


@dataclass(eq=False, repr=False)
class Node(betterproto.Message):
    """A node in a `Graph`"""

    input: "Empty" = betterproto.message_field(1, group="node")
    """The node that emits a graph's input values."""

    output: "Empty" = betterproto.message_field(2, group="node")
    """The node that receives a graph's output values."""

    const: "Value" = betterproto.message_field(3, group="node")
    """A node that emits a constant value (contained within)."""

    box: "BoxNode" = betterproto.message_field(4, group="node")
    """
    A subgraph embedded as a single node. The ports of the node are the ports
     of the embedded graph. Box nodes can be used to conveniently compose common
     subgraphs. The box also specifies a location at which to run the graph.
    """

    function: "FunctionNode" = betterproto.message_field(5, group="node")
    """
    A node that executes a function with a specified name. The type and runtime
     behavior of a function node depend on the functions provided by the environment
     in which the graph is interpreted.
    """

    match: "Empty" = betterproto.message_field(6, group="node")
    """
    Perform pattern matching on a variant value, with handlers according to the
     input `Type::variant`
    """

    tag: str = betterproto.string_field(7, group="node")
    """Create a variant. Tag(tag) :: forall T. T -> Variant(tag:T | ...)"""


@dataclass(eq=False, repr=False)
class Edge(betterproto.Message):
    """An edge in a `Graph`"""

    port_from: str = betterproto.string_field(1)
    """Source (out-)port of `node_from`"""

    port_to: str = betterproto.string_field(2)
    """Target/destination (in-)port of `node_to`"""

    node_from: int = betterproto.uint32_field(3)
    """Source node"""

    node_to: int = betterproto.uint32_field(4)
    """Source port"""

    edge_type: "Type" = betterproto.message_field(5)
    """
    Explicit annotation of the type of the edge. Client may (optionally)
     provide; typechecking will (always) fill in.
    """


@dataclass(eq=False, repr=False)
class Graph(betterproto.Message):
    """
    A computation graph is a directed acyclic port graph in which data flows along
     the edges and as it is processed by the nodes.  Nodes in the graph are densely numbered
     starting from 0. The inputs of the graph are emitted by a `Node::input` at index 0,
     and the outputs are received by a `Node::output` at index 1.
     Each node has labelled input and output ports, encoded implicitly as the endpoints of the
     graph's edges. The port labels are unique among a node's input and output ports individually,
     but input and output ports with the same label are considered different. Any port in the graph
     has exactly one edge connected to it.
    """

    nodes: List["Node"] = betterproto.message_field(1)
    """
    The nodes in the graph. The first two must be a `Node::input` and a `Node::output`,
     and there must not be any other such nodes at any other index.
    """

    edges: List["Edge"] = betterproto.message_field(2)
    """
    The edges in the graph. Each can optionally be annotated with a type.
    """

    name: str = betterproto.string_field(3)
    """
    User-provided name of the graph. Used for debug/display, does not affect execution.
    """

    input_order: List[str] = betterproto.string_field(4)
    """
    Optional ordering of input ports used by graph builders if available.
     Listed ports will be reported and wired up in the order given
     and before ports that are not listed
    """

    output_order: List[str] = betterproto.string_field(5)
    """Like `input_order` but for the output ports."""


@dataclass(eq=False, repr=False)
class Empty(betterproto.Message):
    """
    This is supposed to be `google.protobuf.Empty` but unfortunately there is no
     support for this in `betterproto` yet.
    """

    pass


@dataclass(eq=False, repr=False)
class TypeScheme(betterproto.Message):
    """
    A polymorphic type scheme. Usually (but not necessarily) for a function.
    """

    variables: List["TypeSchemeVar"] = betterproto.message_field(1)
    """
    Variables over which the scheme is polymorphic, each with its `Kind`.
     A concrete type (usable for a port or edge) can be obtained by giving
     a type or row (according to the `Kind`) for each.
    """

    constraints: List["Constraint"] = betterproto.message_field(2)
    """
    Constraints restricting the legal instantiations of the [Self::variables]
    """

    body: "Type" = betterproto.message_field(3)
    """
    The body of the type scheme, i.e. potentially containing occurrences of the
     `variables`, into which values (`Type`s or `RowTypes`) for the variables
     are substituted in order to instantiate the scheme into a concrete type.
    """


@dataclass(eq=False, repr=False)
class TypeSchemeVar(betterproto.Message):
    """Describes a variable bound by a `TypeScheme`"""

    name: str = betterproto.string_field(1)
    """Name of the variable"""

    kind: "Kind" = betterproto.message_field(2)
    """Whether the variable is a `Type` of `Value`s or a `RowType`"""


@dataclass(eq=False, repr=False)
class Constraint(betterproto.Message):
    """
    Specifies a restriction on possible instantiations of type variables in a [TypeScheme]
    """

    lacks: "LacksConstraint" = betterproto.message_field(1, group="constraint")
    """A type variable of `Kind::Row` does *not* have a specified label"""

    partition: "PartitionConstraint" = betterproto.message_field(2, group="constraint")
    """
    A row is the union of two other rows (we expect at least two of the three
     rows to contain variables)
    """


@dataclass(eq=False, repr=False)
class PartitionConstraint(betterproto.Message):
    """
    `Constraint` that a row is the union of two other rows (which for any label in common,
     must have `Type`s that can be made the same)
    """

    left: "Type" = betterproto.message_field(1)
    """
    One input to the union - a `Type::Row` or a `Type::Var` of `Kind::Row`
    """

    right: "Type" = betterproto.message_field(2)
    """
    The other input to the union - a `Type::Row` or a `Type::Var` of `Kind::Row`
    """

    union: "Type" = betterproto.message_field(3)
    """
    The result, i.e. `left` and `right` merged together.
     Could be a `Type::Row` or a `Type::Var` of `Kind::Row`
    """


@dataclass(eq=False, repr=False)
class LacksConstraint(betterproto.Message):
    """
    `Constraint` that a row does not contain an element with the specified name
    """

    row: "Type" = betterproto.message_field(1)
    """A `Type::row` or a `Type::var` of `Kind::row`"""

    label: str = betterproto.string_field(2)
    """Field name that must not be present in `row`"""


@dataclass(eq=False, repr=False)
class Kind(betterproto.Message):
    """
    The kind of a type variable - i.e. whether the "type" variable stands for
     a single type, or (some part of) a row.
    """

    star: "Empty" = betterproto.message_field(1, group="kind")
    """Kind of types (describing sets of values)"""

    row: "Empty" = betterproto.message_field(2, group="kind")
    """Kind of rows (unordered names each with a type)"""



"""Send requests to tierkreis server to execute a graph."""

import asyncio
from collections.abc import Coroutine
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Type,
    TypeVar,
    cast,
)

import betterproto
from grpclib.client import Channel
from grpclib.events import SendRequest

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
import tierkreis.core.protos.tierkreis.v1alpha1.runtime as pr
import tierkreis.core.protos.tierkreis.v1alpha1.signature as ps
from tierkreis.client.runtime_client import RuntimeClient
from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import Location, TierkreisGraph
from tierkreis.core.type_errors import TierkreisTypeErrors
from tierkreis.core.values import IncompatiblePyValue, StructValue, TierkreisValue

if TYPE_CHECKING:
    from betterproto.grpc.grpclib_client import ServiceStub


@dataclass
class RuntimeHTTPError(Exception):
    """Error while communicating with tierkreis server."""

    endpoint: str
    code: int
    content: str

    def __str__(self) -> str:
        return (
            f"Request to endpoint '{self.endpoint}'"
            f" failed with code {self.code}"
            f" and content '{self.content}'."
        )


@dataclass
class InputConversionError(Exception):
    input_name: str
    value: Any

    def __str__(self) -> str:
        return (
            f"Input value with name {self.input_name} cannot be"
            "converted to a TierkreisValue."
        )


StubType = TypeVar("StubType", bound="ServiceStub")


class ServerRuntime(RuntimeClient):
    """Client for a tierkreis server connected over a GRPC channel."""

    def __init__(self, channel: "Channel") -> None:
        """Initialize the client with a GRPC channel to a runtime server."""
        self._channel = channel
        self._stubs: dict[str, ServiceStub] = {}

    def socket_address(self) -> str:
        """Socket address (host and port) of the server."""
        return f"{self._channel._host}:{self._channel._port}"

    def _stub_gen(self, key: str, stub_t: Type[StubType]) -> StubType:
        if key not in self._stubs:
            self._stubs[key] = stub_t(self._channel)
        stub = self._stubs[key]
        assert isinstance(stub, stub_t)
        return stub

    @property
    def _signature_stub(self) -> ps.SignatureStub:
        return self._stub_gen("signature", ps.SignatureStub)

    @property
    def _runtime_stub(self) -> pr.RuntimeStub:
        return self._stub_gen("runtime", pr.RuntimeStub)

    @property
    def _type_stub(self) -> ps.TypeInferenceStub:
        return self._stub_gen("type", ps.TypeInferenceStub)

    async def get_signature(self, loc: Location = Location([])) -> Signature:
        """Retrieve the namespaces and functions available on the runtime."""
        return Signature.from_proto(
            await self._signature_stub.list_functions(ps.ListFunctionsRequest(loc))
        )

    async def run_graph(
        self,
        graph: TierkreisGraph,
        loc: Location = Location([]),
        /,
        **py_inputs: Any,
    ) -> Dict[str, TierkreisValue]:
        inputs = {}
        for key, val in py_inputs.items():
            try:
                inputs[key] = TierkreisValue.from_python(val)
            except IncompatiblePyValue as err:
                raise InputConversionError(key, val) from err

        decoded = await self._runtime_stub.run_graph(
            pr.RunGraphRequest(
                graph=graph.to_proto(),
                inputs=pg.StructValue(map=StructValue(inputs).to_proto_dict()),
                type_check=True,
                loc=loc,
            )
        )

        status, status_value = betterproto.which_one_of(decoded, "result")
        assert status_value is not None
        if status == "type_errors":
            raise TierkreisTypeErrors.from_proto(status_value, graph)
        if status == "error":
            raise RuntimeError(
                f"Run_graph execution failed with message:\n{status_value}"
            )
        return StructValue.from_proto_dict(status_value.map).values

    def run_graph_block(
        self,
        graph: TierkreisGraph,
        /,
        **py_inputs: Any,
    ) -> Dict[str, TierkreisValue]:
        """Blocking version of `run_graph`."""

        async def _run(
            host: str,
            port: int,
        ):
            async with Channel(host, port) as channel:
                return await ServerRuntime(channel).run_graph(graph, **py_inputs)

        return async_to_sync(_run)(self._channel._host, self._channel._port)

    async def type_check_graph(
        self, graph: TierkreisGraph, loc: Location = Location([])
    ) -> TierkreisGraph:
        """Type check the graph on the runtime and return a new type-annotated graph."""
        value = TierkreisValue.from_python(graph).to_proto()

        response = await self._type_stub.infer_type(
            ps.InferTypeRequest(value=value, loc=loc)
        )
        name, message = betterproto.which_one_of(response, "response")

        if name == "success":
            message = cast(ps.InferTypeSuccess, message)
            assert message.value.graph is not None
            return TierkreisValue.from_proto(message.value).to_python(TierkreisGraph)

        errors = cast(ps.TypeErrors, message)
        raise TierkreisTypeErrors.from_proto(errors, graph)


def _gen_auth_injector(login: str, pwd: str) -> Callable[["SendRequest"], Coroutine]:
    async def _inject_auth(event: SendRequest) -> None:
        event.metadata["token"] = login
        event.metadata["key"] = pwd

    return _inject_auth


class RuntimeLaunchFailed(Exception):
    """Starting server locally failed."""


CallableReturn = TypeVar("CallableReturn")


def async_to_sync(
    func: Callable[..., Coroutine[None, None, CallableReturn]],
) -> Callable[..., CallableReturn]:
    """
    Converts an asynchronous function into a synchronous one by running it
    on a new async event loop in a newly created thread.
    """

    @wraps(func)
    def sync(*args, **kwargs):
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(lambda: asyncio.run(func(*args, **kwargs)))
            return future.result()

    return sync



"""Worker server implementation."""

import functools
import sys
from contextvars import ContextVar
from tempfile import TemporaryDirectory
from traceback import print_exception
from typing import Any, Callable, Coroutine, Optional

import grpclib
import grpclib.events
import grpclib.server
from grpclib.const import Status as StatusCode
from grpclib.exceptions import GRPCError
from grpclib.server import Server

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
import tierkreis.core.protos.tierkreis.v1alpha1.runtime as pr
import tierkreis.core.protos.tierkreis.v1alpha1.signature as ps
import tierkreis.core.protos.tierkreis.v1alpha1.worker as pw
from tierkreis.core.function import FunctionName
from tierkreis.core.protos.tierkreis.v1alpha1.worker import (
    RunFunctionResponse,
    WorkerBase,
)
from tierkreis.core.tierkreis_graph import TierkreisGraph
from tierkreis.core.type_errors import TierkreisTypeErrors
from tierkreis.core.values import StructValue
from tierkreis.pyruntime.python_runtime import PyRuntime
from tierkreis.worker.callback import callback_server

from .exceptions import (
    DecodeInputError,
    EncodeOutputError,
    FunctionNotFound,
    NodeExecutionError,
)
from .namespace import Metadata, Namespace
from .tracing import _TRACING, context_token, get_tracer, span

tracer = get_tracer(__name__)


async def _event_recv_request(request: grpclib.events.RecvRequest):
    method_func = request.method_func
    service, method = request.method_name.lstrip("/").split("/", 1)
    kwargs: dict[str, Any] = dict(name=f"GRPC: {request.method_name}")
    if _TRACING:
        import opentelemetry.propagate
        import opentelemetry.trace
        from opentelemetry.semconv.trace import SpanAttributes

        context = opentelemetry.propagate.extract(request.metadata)
        attributes = {
            SpanAttributes.RPC_SYSTEM: "grpc",
            SpanAttributes.RPC_SERVICE: service,
            SpanAttributes.RPC_METHOD: method,
        }
        kwargs["kind"] = opentelemetry.trace.SpanKind.SERVER
        kwargs["attributes"] = attributes
    else:
        context = None

    async def wrapped(stream: grpclib.server.Stream):
        with context_token(context):
            with span(tracer, **kwargs) as s:
                from .prelude import profile_worker  # avoid cyclic import

                if profile_worker:
                    import cProfile  # could be global
                    import pstats
                    from io import StringIO

                    render_results = StringIO()
                    with cProfile.Profile() as pr:
                        # If there are multiple requests being processed at the
                        # same time, due to "async" this will likely get confused,
                        # i.e. combine processing for any+all of them together.
                        pr.enable()
                        await method_func(stream)
                        pr.disable()
                        stats = pstats.Stats(pr, stream=render_results)
                    stats.sort_stats("cumtime")
                    stats.print_stats(20)  # Top 20 lines
                    render_results.seek(0)
                    s.set_attribute("profile_results", render_results.read())
                else:
                    await method_func(stream)

    request.method_func = wrapped


_KEYRING_SERVICE = "tierkreis_extracted"


class Worker:
    """Worker server."""

    root: Namespace
    server: Server
    pyruntime: PyRuntime
    metadata: ContextVar[Metadata]

    def __init__(self, root_namespace):
        self.root = root_namespace
        self.pyruntime = PyRuntime([root_namespace])
        self.server = Server(
            [SignatureServerImpl(self), WorkerServerImpl(self), RuntimeServerImpl(self)]
        )
        self.metadata = ContextVar("metadata")

        # Attach event listener for tracing
        self._add_request_listener(_event_recv_request)
        # Attach event listener to extract metadata
        self._add_request_listener(self._record_metadata)

    async def run(
        self,
        function: FunctionName,
        inputs: StructValue,
        callback: pr.Callback,
        metadata: Metadata,
    ) -> StructValue:
        """Run function."""
        func = self.root.get_function(function)
        if func is None:
            raise FunctionNotFound(function)

        async with callback_server(callback) as cb:
            return await func.run(cb, metadata, inputs)

    async def _record_metadata(self, request: grpclib.events.RecvRequest) -> None:
        method_func = request.method_func

        @functools.wraps(method_func)
        async def wrapped(stream: grpclib.server.Stream):
            token = self.metadata.set(request.metadata)
            await method_func(stream)
            # Good practice but probably not essential:
            self.metadata.reset(token)

        request.method_func = wrapped

    def _add_request_listener(
        self,
        listener: Callable[[grpclib.events.RecvRequest], Coroutine[Any, Any, None]],
    ):
        grpclib.events.listen(self.server, grpclib.events.RecvRequest, listener)

    async def start(self, port: Optional[int] = None):
        """Start server."""

        if port:
            await self.server.start(port=port)

            print(f"Started worker server on port: {port}", flush=True)
            await self.server.wait_closed()

        else:
            with TemporaryDirectory() as socket_dir:
                # Create a temporary path for a unix domain socket.
                socket_path = f"{socket_dir}/python_worker.sock"

                # Start the python worker gRPC server and bind to the unix domain socket
                await self.server.start(path=socket_path)

                # Print the path of the unix domain socket to stdout so the runtime can
                # connect to it. Without the flush the runtime did not receive the
                # socket path and blocked indefinitely.
                print(socket_path)
                sys.stdout.flush()

                await self.server.wait_closed()


class WorkerServerImpl(WorkerBase):
    worker: Worker

    def __init__(self, worker: Worker):
        self.worker = worker

    async def run_function(
        self, run_function_request: pw.RunFunctionRequest
    ) -> RunFunctionResponse:
        function = run_function_request.function
        inputs = run_function_request.inputs
        callback = run_function_request.callback
        metadata = self.worker.metadata.get()
        try:
            function_name = FunctionName.from_proto(function)
            inputs_struct = StructValue.from_proto_dict(inputs.map)
            outputs_struct = await self.worker.run(
                function_name, inputs_struct, callback, metadata
            )
            with span(tracer, name="encoding python type in RunFunctionResponse proto"):
                res = RunFunctionResponse(
                    outputs=pg.StructValue(outputs_struct.to_proto_dict())
                )
            return res
        except DecodeInputError as err:
            raise GRPCError(
                status=StatusCode.INVALID_ARGUMENT,
                message=f"Error while decoding inputs: {err}",
            ) from err
        except EncodeOutputError as err:
            raise GRPCError(
                status=StatusCode.INTERNAL,
                message=f"Error while encoding outputs: {err}",
            ) from err
        except FunctionNotFound as err:
            raise GRPCError(
                status=StatusCode.UNIMPLEMENTED,
                message=f"Unsupported function: {function}",
            ) from err
        except NodeExecutionError as err:
            # The response resulting from the GRPCError below does not include
            # the original traceback to avoid leaking implementation details.
            # The traceback is instead printed to local stderr.
            print(f"Error in running {function}:", file=sys.stderr)
            print_exception(err.base_exception)

            raise GRPCError(
                status=StatusCode.UNKNOWN,
                message=f"Error while running operation: {repr(err.base_exception)}",
            ) from err


class SignatureServerImpl(ps.SignatureBase):
    worker: Worker

    def __init__(self, worker: Worker):
        self.worker = worker

    async def list_functions(
        self, list_functions_request: ps.ListFunctionsRequest
    ) -> ps.ListFunctionsResponse:
        signature = self.worker.root.extract_signature(True)

        return signature.to_proto()


class RuntimeServerImpl(pr.RuntimeBase):
    worker: Worker

    def __init__(self, worker: Worker):
        self.worker = worker

    async def run_graph(
        self, run_graph_request: pr.RunGraphRequest
    ) -> pr.RunGraphResponse:
        graph = TierkreisGraph.from_proto(run_graph_request.graph)
        inputs = StructValue.from_proto_dict(run_graph_request.inputs.map)
        try:
            if run_graph_request.type_check:
                (
                    graph,
                    inputs,
                ) = await self.worker.pyruntime.type_check_graph_with_inputs(
                    graph, inputs
                )

            outputs = await self.worker.pyruntime.run_graph(graph, **inputs.values)
            return pr.RunGraphResponse(
                success=pg.StructValue(map=StructValue(outputs).to_proto_dict())
            )
        except TierkreisTypeErrors as err:
            return pr.RunGraphResponse(type_errors=err.to_proto())
        except Exception as err:
            return pr.RunGraphResponse(error=str(err))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: v1alpha1/signature.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import graph as _graph__


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class FunctionDeclaration(betterproto.Message):
    """
    Information about a function: its polymorphic [TypeScheme], user-readable
     description, and port ordering (for debug/display purposes, not used by
     the typechecker).
    """

    type_scheme: "_graph__.TypeScheme" = betterproto.message_field(2)
    """/ Polymorphic type scheme describing all possible input/output types"""

    description: str = betterproto.string_field(3)
    """Human-readable documentation"""

    input_order: List[str] = betterproto.string_field(4)
    """Order in which to display input ports"""

    output_order: List[str] = betterproto.string_field(5)
    """Order in which to display output ports"""


@dataclass(eq=False, repr=False)
class ListFunctionsRequest(betterproto.Message):
    """
    Request to list the functions known to a Runtime, optionally filtering by a Location
    """

    loc: "_graph__.Location" = betterproto.message_field(1)
    """
    Filter to only report functions available in the specified location or children
     thereof. (The default/empty location means the root, i.e. all functions.)
    """


@dataclass(eq=False, repr=False)
class ListFunctionsResponse(betterproto.Message):
    """
    aka the "Signature" of a Runtime: the `FunctionDeclaration`s and
     descendant `Location`s that it knows.
    """

    root: "Namespace" = betterproto.message_field(1)
    """
    Every function the runtime can run, and the locations in which it can run each
    """

    aliases: Dict[str, "_graph__.TypeScheme"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """Named aliases for polymorphic types"""

    scopes: List["_graph__.Location"] = betterproto.message_field(3)
    """All the locations the Runtime knows."""


@dataclass(eq=False, repr=False)
class InferTypeRequest(betterproto.Message):
    """
    Request to infer the type of a graph (used for `TypeInference::InferType`)
    """

    value: "_graph__.Value" = betterproto.message_field(1)
    """Value whose type (scheme) to infer"""

    loc: "_graph__.Location" = betterproto.message_field(2)
    """
    Sub-location of the runtime in to which to check. If this leads to
     `ErrorVariant::unknown_function` errors that do not occur without the
     location, this indicates usage of functions available only outside that
     location i.e. these would use an escape hatch if the Graph were run.
    """


@dataclass(eq=False, repr=False)
class InferTypeResponse(betterproto.Message):
    """
    Result of inferring the type of a value via `TypeInference::InferType`
    """

    success: "InferTypeSuccess" = betterproto.message_field(1, group="response")
    """
    Inference succeeded, return the type-annotated value and inferred type (scheme)
    """

    error: "TypeErrors" = betterproto.message_field(2, group="response")
    """Type inference failed because of one or more type errors"""


@dataclass(eq=False, repr=False)
class InferTypeSuccess(betterproto.Message):
    """A type successfully inferred by `TypeInference::InferType`"""

    value: "_graph__.Value" = betterproto.message_field(1)
    """
    The value whose type was inferred. The same as the value passed in,
     except that any `Value::graph`s within will have their edges annotated
     with the inferred types (`Edge::edge_type`).
    """

    type_scheme: "_graph__.TypeScheme" = betterproto.message_field(2)
    """
    Type scheme inferred for the value, i.e. explicitly listing any type variables
     over which the value is polymorphic. (E.g. if the value is an empty list.)
    """


@dataclass(eq=False, repr=False)
class GraphWithInputs(betterproto.Message):
    """
    A graph with (optionally) input values for it. Used for
     `InferGraphTypesRequest`s and `InferGraphTypesResponse`.
    """

    graph: "_graph__.Graph" = betterproto.message_field(1)
    """The graph"""

    inputs: Optional["_graph__.StructValue"] = betterproto.message_field(
        2, optional=True
    )
    """Optionally, input values to feed to the graph"""


@dataclass(eq=False, repr=False)
class InferGraphTypesRequest(betterproto.Message):
    """
    Used by `tierkreis-typecheck` Rust(PYO3)/python interop library
     to request type inference of a graph with input values.
    """

    gwi: "GraphWithInputs" = betterproto.message_field(1)
    """Graph and inputs whose types to infer"""

    functions: "Namespace" = betterproto.message_field(2)
    """The signature of functions to check against."""


@dataclass(eq=False, repr=False)
class InferGraphTypesResponse(betterproto.Message):
    """
    Used by `tierkreis-typecheck` Rust(PYO3)/python interop library
     to report results of type inference on a graph with input values.
    """

    success: "GraphWithInputs" = betterproto.message_field(1, group="response")
    """Inference was successful, return the type"""

    error: "TypeErrors" = betterproto.message_field(2, group="response")
    """Inference failed due to one or more `TierkreisTypeError`s"""


@dataclass(eq=False, repr=False)
class Empty(betterproto.Message):
    """
    This is supposed to be `google.protobuf.Empty` but unfortunately there is no
     support for this in `betterproto` yet.
    """

    pass


@dataclass(eq=False, repr=False)
class GraphLocation(betterproto.Message):
    """
    A series of these identifies where in a `Value::graph` or `GraphWithInputs`
     was the cause of a `TierkreisTypeError`.
    """

    vec_index: int = betterproto.uint32_field(2, group="location")
    """
    Where the previous location(s) identify a `Value::vec`, identifies one element by index
    """

    node_idx: int = betterproto.uint32_field(4, group="location")
    """
    Where the previous location(s) identify a graph, the error is in the indexed node.
     (If-and-only-if the node is a Box or Const, may be followed by more `GraphLocation`s.)
    """

    edge: "_graph__.Edge" = betterproto.message_field(5, group="location")
    """
    Where the previous location(s) identify a graph, the error is on the indicated edge.
     (Will end the sequence of `GraphLocation`s.)
    """

    input: "Empty" = betterproto.message_field(6, group="location")
    """
    Where previous location(s) identify a graph, the error is in the input node
    """

    output: "Empty" = betterproto.message_field(7, group="location")
    """
    Where previous location(s) identify a graph, the error is in the output node
    """

    struct_field: str = betterproto.string_field(8, group="location")
    """
    Where the previous location(s) identify a struct value (e.g. inside a Const),
     the error is in the named field of that struct
    """

    pair_first: "Empty" = betterproto.message_field(9, group="location")
    """
    Where the previous location(s) identify a `Value::pair`, the error is in the first element
    """

    pair_second: "Empty" = betterproto.message_field(10, group="location")
    """
    Where the previous location(s) identify a `Value::pair`, the error is in the second element
    """

    map_key: "Empty" = betterproto.message_field(11, group="location")
    """
    Where the previous location(s) identify a `Value::map`, the error is in one of the keys
     (does not specify which)
    """

    map_value: "Empty" = betterproto.message_field(12, group="location")
    """
    Where the previous location(s) identify a `Value::map`, the error is in one of the values
     (does not specify which)
    """

    input_value: str = betterproto.string_field(13, group="location")
    """
    For `TypeInference::InferType`,
     indicates the error is in one of the `GraphWithInputs::inputs`
    """


@dataclass(eq=False, repr=False)
class TypeErrors(betterproto.Message):
    """A Collection of `TypeError`s"""

    errors: List["TierkreisTypeError"] = betterproto.message_field(1)
    """List of errors"""


@dataclass(eq=False, repr=False)
class UnifyError(betterproto.Message):
    """`ErrorVariant` that two types failed to unify."""

    expected: "_graph__.Type" = betterproto.message_field(1)
    """The type that was expected."""

    found: "_graph__.Type" = betterproto.message_field(2)
    """The type that was actually inferred."""


@dataclass(eq=False, repr=False)
class TypeVarError(betterproto.Message):
    """
    `ErrorVariant` that a type scheme is ill-formed because it refers to an unknown
     type variable.
    """

    variable: "_graph__.TypeSchemeVar" = betterproto.message_field(1)
    """The unknown type variable."""

    type_scheme: "_graph__.TypeScheme" = betterproto.message_field(2)
    """The ill-formed type scheme."""


@dataclass(eq=False, repr=False)
class ErrorVariant(betterproto.Message):
    """Errors that can occur during type checking."""

    unify: "UnifyError" = betterproto.message_field(1, group="error")
    """Two types failed to unify."""

    kind: str = betterproto.string_field(2, group="error")
    """A type scheme is ill-formed due to a kind mismatch."""

    unknown_function: "_graph__.FunctionName" = betterproto.message_field(
        3, group="error"
    )
    """A graph referred to an unknown function."""

    unknown_type_var: "TypeVarError" = betterproto.message_field(4, group="error")
    """
    A type scheme is ill-formed because it refers to an unknown type variable.
    """

    bound: str = betterproto.string_field(5, group="error")
    """
    A type constraint (`LacksConstraint` or `PartitionConstraint`) is unsatisfiable.
    """


@dataclass(eq=False, repr=False)
class TierkreisTypeError(betterproto.Message):
    """
    An error preventing type inference in a graph passed to `TypeInference::InferType`
     or `tierkreis.v1alpha1.runtime.Runime.RunGraph`
    """

    variant: "ErrorVariant" = betterproto.message_field(1)
    """Detail of the error"""

    location: List["GraphLocation"] = betterproto.message_field(2)
    """
    Identifies where in the value/graph the error occurred.
     Locations go from outermost to innermost in nested const/box graphs.
    """


@dataclass(eq=False, repr=False)
class NamespaceItem(betterproto.Message):
    """
    A `FunctionDeclaration` with a set of `Location`s at which
     the function is supported.
    """

    decl: "FunctionDeclaration" = betterproto.message_field(1)
    """
    Declaration of the function, including typescheme;
     identical at all `locations`
    """

    locations: List["_graph__.Location"] = betterproto.message_field(2)
    """The locations (aka scopes) at which the function is supported"""


@dataclass(eq=False, repr=False)
class Namespace(betterproto.Message):
    """
    Tree-structured mapping (sharing common prefixes) from
     `tierkreis.v1alpha1.graph.FunctionName`s to `NamespaceItem`s
    """

    functions: Dict[str, "NamespaceItem"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """`NamespaceItem`s at this level of the qualified-name hierarchy"""

    subspaces: Dict[str, "Namespace"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """
    Mappings for subtrees of the name hierarchy, i.e. for qualnames
     with a longer prefix (the map key being the next atom of prefix)
    """


class SignatureStub(betterproto.ServiceStub):
    async def list_functions(
        self,
        list_functions_request: "ListFunctionsRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ListFunctionsResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.signature.Signature/ListFunctions",
            list_functions_request,
            ListFunctionsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class TypeInferenceStub(betterproto.ServiceStub):
    async def infer_type(
        self,
        infer_type_request: "InferTypeRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "InferTypeResponse":
        return await self._unary_unary(
            "/tierkreis.v1alpha1.signature.TypeInference/InferType",
            infer_type_request,
            InferTypeResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class SignatureBase(ServiceBase):

    async def list_functions(
        self, list_functions_request: "ListFunctionsRequest"
    ) -> "ListFunctionsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_list_functions(
        self,
        stream: "grpclib.server.Stream[ListFunctionsRequest, ListFunctionsResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.list_functions(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha1.signature.Signature/ListFunctions": grpclib.const.Handler(
                self.__rpc_list_functions,
                grpclib.const.Cardinality.UNARY_UNARY,
                ListFunctionsRequest,
                ListFunctionsResponse,
            ),
        }


class TypeInferenceBase(ServiceBase):

    async def infer_type(
        self, infer_type_request: "InferTypeRequest"
    ) -> "InferTypeResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_infer_type(
        self, stream: "grpclib.server.Stream[InferTypeRequest, InferTypeResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.infer_type(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/tierkreis.v1alpha1.signature.TypeInference/InferType": grpclib.const.Handler(
                self.__rpc_infer_type,
                grpclib.const.Cardinality.UNARY_UNARY,
                InferTypeRequest,
                InferTypeResponse,
            ),
        }



"""Prelude definitions and functions for server executable scripts."""

import argparse
import asyncio
import os

from .namespace import Namespace
from .tracing import _TRACING
from .worker import Worker

parser = argparse.ArgumentParser(description="Parse worker server cli.")
parser.add_argument(
    "--port", help="If specified listen on network port rather than UDP."
)

profile_worker: bool


def setup_tracing(service_name: str):
    """If the TIERKREIS_OTLP environment variable is set, set up tracing using opentelemetry."""
    endpoint = os.environ.get("TIERKREIS_OTLP")
    global profile_worker
    profile_worker = bool(os.environ.get("TIERKREIS_PROFILE_WORKER"))

    if endpoint is None:
        return
    if _TRACING:
        import opentelemetry.trace
        from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
            OTLPSpanExporter,
        )
        from opentelemetry.sdk.resources import SERVICE_NAME, Resource
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.sdk.trace.export import SimpleSpanProcessor

        tracer_provider = TracerProvider(
            resource=Resource.create({SERVICE_NAME: service_name})
        )

        span_exporter = OTLPSpanExporter(endpoint=endpoint)

        tracer_provider.add_span_processor(SimpleSpanProcessor(span_exporter))

        opentelemetry.trace.set_tracer_provider(tracer_provider)


def start_worker_server(worker_name: str, namespace: Namespace):
    """Set up tracing and run the worker server with the provided namespaces.
    Expects a port specified on the command line, and reports successful start to
    stdout
    """

    async def main():
        args = parser.parse_args()
        setup_tracing(worker_name)
        worker = Worker(namespace)
        await worker.start(args.port)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())
